﻿Option Explicit On
Module _6502
    ' Opcode Constants
    '---------------------------------------
    Public Const INS_ADC As Long = 0
    Public Const INS_AND As Long = 1
    Public Const INS_ASL As Long = 2
    Public Const INS_BCC As Long = 3
    Public Const INS_BCS As Long = 4
    Public Const INS_BEQ As Long = 5
    Public Const INS_BIT As Long = 6
    Public Const INS_BMI As Long = 7
    Public Const INS_BNE As Long = 8
    Public Const INS_BPL As Long = 9
    Public Const INS_BRK As Long = 10
    Public Const INS_BVC As Long = 11
    Public Const INS_BVS As Long = 12
    Public Const INS_CLC As Long = 13
    Public Const INS_CLD As Long = 14
    Public Const INS_CLI As Long = 15
    Public Const INS_CLV As Long = 16
    Public Const INS_CMP As Long = 17
    Public Const INS_CPX As Long = 18
    Public Const INS_CPY As Long = 19
    Public Const INS_DEC As Long = 20
    Public Const INS_DEX As Long = 21
    Public Const INS_DEY As Long = 22
    Public Const INS_EOR As Long = 23
    Public Const INS_INC As Long = 24
    Public Const INS_INX As Long = 25
    Public Const INS_INY As Long = 26
    Public Const INS_JMP As Long = 27
    Public Const INS_JSR As Long = 28
    Public Const INS_LDA As Long = 29
    Public Const INS_LDX As Long = 30
    Public Const INS_LDY As Long = 31
    Public Const INS_LSR As Long = 32
    Public Const INS_NOP As Long = 33
    Public Const INS_ORA As Long = 34
    Public Const INS_PHA As Long = 35
    Public Const INS_PHP As Long = 36
    Public Const INS_PLA As Long = 37
    Public Const INS_PLP As Long = 38
    Public Const INS_ROL As Long = 39
    Public Const INS_ROR As Long = 40
    Public Const INS_RTI As Long = 41
    Public Const INS_RTS As Long = 42
    Public Const INS_SBC As Long = 43
    Public Const INS_SEC As Long = 44
    Public Const INS_SED As Long = 45
    Public Const INS_SEI As Long = 46
    Public Const INS_STA As Long = 47
    Public Const INS_STX As Long = 48
    Public Const INS_STY As Long = 49
    Public Const INS_TAX As Long = 50
    Public Const INS_TAY As Long = 51
    Public Const INS_TSX As Long = 52
    Public Const INS_TXA As Long = 53
    Public Const INS_TXS As Long = 54
    Public Const INS_TYA As Long = 55

    ' Addressing Mode Constants
    '---------------------------------------
    Public Const ADDRESS_IMPLIED As Long = 0
    Public Const ADDRESS_ACCUMULATOR As Long = 1
    Public Const ADDRESS_IMMEDIATE As Long = 2
    Public Const ADDRESS_ZERO_PAGE As Long = 3
    Public Const ADDRESS_ZERO_PAGE_X As Long = 4
    Public Const ADDRESS_ZERO_PAGE_Y As Long = 5
    Public Const ADDRESS_RELATIVE As Long = 6
    Public Const ADDRESS_ABSOLUTE As Long = 7
    Public Const ADDRESS_ABSOLUTE_X As Long = 8
    Public Const ADDRESS_ABSOLUTE_Y As Long = 9
    Public Const ADDRESS_INDIRECT As Long = 10
    Public Const ADDRESS_INDIRECT_X As Long = 11
    Public Const ADDRESS_INDIRECT_Y As Long = 12

    'Registers
    Private a As Byte
    Private X As Byte
    Private Y As Byte
    Private s As Byte
    Private P As Byte

    Private PC As Integer
    Private SavePC As Integer
    Private Value As Integer
    Private Sum As Integer
    Private SaveFlags As Integer
    Private Help As Integer

    Public Opcode As Byte
    Public Tick_Count As Integer

    'Arrays
    Public Ticks(0 To &H100&) As Byte
    Public AddrMode(0 To &H100&) As Byte
    Public Instruction(0 To &H100&) As Byte

    'Misc
    Private ExecutedCycles As Integer
    Public CurrentLine As Integer
    Const MaxCycles As Integer = 113
    Public Sub Initialize_CPU()
        Ticks(&H0) = 7 : Instruction(&H0) = INS_BRK : AddrMode(&H0) = ADDRESS_RELATIVE
        Ticks(&H1) = 6 : Instruction(&H1) = INS_ORA : AddrMode(&H1) = ADDRESS_INDIRECT_X
        Ticks(&H2) = 2 : Instruction(&H2) = INS_NOP : AddrMode(&H2) = ADDRESS_IMPLIED
        Ticks(&H3) = 8 : Instruction(&H3) = INS_NOP : AddrMode(&H3) = ADDRESS_IMPLIED
        Ticks(&H4) = 3 : Instruction(&H4) = INS_NOP : AddrMode(&H4) = ADDRESS_IMPLIED
        Ticks(&H5) = 3 : Instruction(&H5) = INS_ORA : AddrMode(&H5) = ADDRESS_ZERO_PAGE
        Ticks(&H6) = 5 : Instruction(&H6) = INS_ASL : AddrMode(&H6) = ADDRESS_ZERO_PAGE
        Ticks(&H7) = 5 : Instruction(&H7) = INS_NOP : AddrMode(&H7) = ADDRESS_IMPLIED
        Ticks(&H8) = 3 : Instruction(&H8) = INS_PHP : AddrMode(&H8) = ADDRESS_IMPLIED
        Ticks(&H9) = 2 : Instruction(&H9) = INS_ORA : AddrMode(&H9) = ADDRESS_IMMEDIATE
        Ticks(&HA) = 2 : Instruction(&HA) = INS_ASL : AddrMode(&HA) = ADDRESS_ACCUMULATOR
        Ticks(&HB) = 2 : Instruction(&HB) = INS_NOP : AddrMode(&HB) = ADDRESS_IMPLIED
        Ticks(&HC) = 4 : Instruction(&HC) = INS_NOP : AddrMode(&HC) = ADDRESS_IMPLIED
        Ticks(&HD) = 4 : Instruction(&HD) = INS_ORA : AddrMode(&HD) = ADDRESS_ABSOLUTE
        Ticks(&HE) = 6 : Instruction(&HE) = INS_ASL : AddrMode(&HE) = ADDRESS_ABSOLUTE
        Ticks(&HF) = 6 : Instruction(&HF) = INS_NOP : AddrMode(&HF) = ADDRESS_IMPLIED

        Ticks(&H10) = 2 : Instruction(&H10) = INS_BPL : AddrMode(&H10) = ADDRESS_RELATIVE
        Ticks(&H11) = 5 : Instruction(&H11) = INS_ORA : AddrMode(&H11) = ADDRESS_INDIRECT_Y       'CORRECT
        Ticks(&H12) = 2 : Instruction(&H12) = INS_NOP : AddrMode(&H12) = ADDRESS_IMPLIED
        Ticks(&H13) = 8 : Instruction(&H13) = INS_NOP : AddrMode(&H13) = ADDRESS_IMPLIED
        Ticks(&H14) = 4 : Instruction(&H14) = INS_NOP : AddrMode(&H14) = ADDRESS_IMPLIED
        Ticks(&H15) = 4 : Instruction(&H15) = INS_ORA : AddrMode(&H15) = ADDRESS_ZERO_PAGE_X      'CORRECT
        Ticks(&H16) = 6 : Instruction(&H16) = INS_ASL : AddrMode(&H16) = ADDRESS_ZERO_PAGE_X      'CORRECT
        Ticks(&H17) = 6 : Instruction(&H17) = INS_NOP : AddrMode(&H17) = ADDRESS_IMPLIED
        Ticks(&H18) = 2 : Instruction(&H18) = INS_CLC : AddrMode(&H18) = ADDRESS_IMPLIED
        Ticks(&H19) = 4 : Instruction(&H19) = INS_ORA : AddrMode(&H19) = ADDRESS_ABSOLUTE_Y       'CORRECT
        Ticks(&H1A) = 2 : Instruction(&H1A) = INS_NOP : AddrMode(&H1A) = ADDRESS_IMPLIED
        Ticks(&H1B) = 7 : Instruction(&H1B) = INS_NOP : AddrMode(&H1B) = ADDRESS_IMPLIED
        Ticks(&H1C) = 4 : Instruction(&H1C) = INS_NOP : AddrMode(&H1C) = ADDRESS_IMPLIED
        Ticks(&H1D) = 4 : Instruction(&H1D) = INS_ORA : AddrMode(&H1D) = ADDRESS_ABSOLUTE_X       'CORRECT
        Ticks(&H1E) = 7 : Instruction(&H1E) = INS_ASL : AddrMode(&H1E) = ADDRESS_ABSOLUTE_X       'CORRECT
        Ticks(&H1F) = 7 : Instruction(&H1F) = INS_NOP : AddrMode(&H1F) = ADDRESS_IMPLIED

        Ticks(&H20) = 6 : Instruction(&H20) = INS_JSR : AddrMode(&H20) = ADDRESS_ABSOLUTE
        Ticks(&H21) = 6 : Instruction(&H21) = INS_AND : AddrMode(&H21) = ADDRESS_INDIRECT_X       'CORRECT
        Ticks(&H22) = 2 : Instruction(&H22) = INS_NOP : AddrMode(&H22) = ADDRESS_IMPLIED
        Ticks(&H23) = 8 : Instruction(&H23) = INS_NOP : AddrMode(&H23) = ADDRESS_IMPLIED
        Ticks(&H24) = 3 : Instruction(&H24) = INS_BIT : AddrMode(&H24) = ADDRESS_ZERO_PAGE        'CORRECT
        Ticks(&H25) = 3 : Instruction(&H25) = INS_AND : AddrMode(&H25) = ADDRESS_ZERO_PAGE        'CORRECT
        Ticks(&H26) = 5 : Instruction(&H26) = INS_ROL : AddrMode(&H26) = ADDRESS_ZERO_PAGE        'CORRECT
        Ticks(&H27) = 5 : Instruction(&H27) = INS_NOP : AddrMode(&H27) = ADDRESS_IMPLIED
        Ticks(&H28) = 4 : Instruction(&H28) = INS_PLP : AddrMode(&H28) = ADDRESS_IMPLIED
        Ticks(&H29) = 2 : Instruction(&H29) = INS_AND : AddrMode(&H29) = ADDRESS_IMMEDIATE        'CORRECT
        Ticks(&H2A) = 2 : Instruction(&H2A) = INS_ROL : AddrMode(&H2A) = ADDRESS_ACCUMULATOR      'CORRECT
        Ticks(&H2B) = 2 : Instruction(&H2B) = INS_NOP : AddrMode(&H2B) = ADDRESS_IMPLIED
        Ticks(&H2C) = 4 : Instruction(&H2C) = INS_BIT : AddrMode(&H2C) = ADDRESS_ABSOLUTE
        Ticks(&H2D) = 4 : Instruction(&H2D) = INS_AND : AddrMode(&H2D) = ADDRESS_ABSOLUTE
        Ticks(&H2E) = 6 : Instruction(&H2E) = INS_ROL : AddrMode(&H2E) = ADDRESS_ABSOLUTE
        Ticks(&H2F) = 6 : Instruction(&H2F) = INS_NOP : AddrMode(&H2F) = ADDRESS_IMPLIED

        Ticks(&H30) = 2 : Instruction(&H30) = INS_BMI : AddrMode(&H30) = ADDRESS_RELATIVE        'CORRECT
        Ticks(&H31) = 5 : Instruction(&H31) = INS_AND : AddrMode(&H31) = ADDRESS_INDIRECT_Y      'CORRECT
        Ticks(&H32) = 2 : Instruction(&H32) = INS_NOP : AddrMode(&H32) = ADDRESS_IMPLIED
        Ticks(&H33) = 8 : Instruction(&H33) = INS_NOP : AddrMode(&H33) = ADDRESS_IMPLIED
        Ticks(&H34) = 4 : Instruction(&H34) = INS_NOP : AddrMode(&H34) = ADDRESS_IMPLIED
        Ticks(&H35) = 4 : Instruction(&H35) = INS_AND : AddrMode(&H35) = ADDRESS_ZERO_PAGE_X
        Ticks(&H36) = 6 : Instruction(&H36) = INS_ROL : AddrMode(&H36) = ADDRESS_ZERO_PAGE_X
        Ticks(&H37) = 6 : Instruction(&H37) = INS_NOP : AddrMode(&H37) = ADDRESS_IMPLIED
        Ticks(&H38) = 2 : Instruction(&H38) = INS_SEC : AddrMode(&H38) = ADDRESS_IMPLIED
        Ticks(&H39) = 4 : Instruction(&H39) = INS_AND : AddrMode(&H39) = ADDRESS_ABSOLUTE_Y
        Ticks(&H3A) = 2 : Instruction(&H3A) = INS_NOP : AddrMode(&H3A) = ADDRESS_IMPLIED
        Ticks(&H3B) = 7 : Instruction(&H3B) = INS_NOP : AddrMode(&H3B) = ADDRESS_IMPLIED
        Ticks(&H3C) = 4 : Instruction(&H3C) = INS_BIT : AddrMode(&H3C) = ADDRESS_ABSOLUTE_X
        Ticks(&H3D) = 4 : Instruction(&H3D) = INS_AND : AddrMode(&H3D) = ADDRESS_ABSOLUTE_X
        Ticks(&H3E) = 7 : Instruction(&H3E) = INS_ROL : AddrMode(&H3E) = ADDRESS_ABSOLUTE_X
        Ticks(&H3F) = 7 : Instruction(&H3F) = INS_NOP : AddrMode(&H3F) = ADDRESS_IMPLIED

        Ticks(&H40) = 6 : Instruction(&H40) = INS_RTI : AddrMode(&H40) = ADDRESS_IMPLIED
        Ticks(&H41) = 6 : Instruction(&H41) = INS_EOR : AddrMode(&H41) = ADDRESS_INDIRECT_X
        Ticks(&H42) = 2 : Instruction(&H42) = INS_NOP : AddrMode(&H42) = ADDRESS_IMPLIED
        Ticks(&H43) = 8 : Instruction(&H43) = INS_NOP : AddrMode(&H43) = ADDRESS_IMPLIED
        Ticks(&H44) = 3 : Instruction(&H44) = INS_NOP : AddrMode(&H44) = ADDRESS_IMPLIED
        Ticks(&H45) = 3 : Instruction(&H45) = INS_EOR : AddrMode(&H45) = ADDRESS_ZERO_PAGE
        Ticks(&H46) = 5 : Instruction(&H46) = INS_LSR : AddrMode(&H46) = ADDRESS_ZERO_PAGE
        Ticks(&H47) = 5 : Instruction(&H47) = INS_NOP : AddrMode(&H47) = ADDRESS_IMPLIED
        Ticks(&H48) = 3 : Instruction(&H48) = INS_PHA : AddrMode(&H48) = ADDRESS_IMPLIED
        Ticks(&H49) = 2 : Instruction(&H49) = INS_EOR : AddrMode(&H49) = ADDRESS_IMMEDIATE
        Ticks(&H4A) = 2 : Instruction(&H4A) = INS_LSR : AddrMode(&H4A) = ADDRESS_ACCUMULATOR
        Ticks(&H4B) = 2 : Instruction(&H4B) = INS_NOP : AddrMode(&H4B) = ADDRESS_IMPLIED
        Ticks(&H4C) = 3 : Instruction(&H4C) = INS_JMP : AddrMode(&H4C) = ADDRESS_ABSOLUTE
        Ticks(&H4D) = 4 : Instruction(&H4D) = INS_EOR : AddrMode(&H4D) = ADDRESS_ABSOLUTE
        Ticks(&H4E) = 6 : Instruction(&H4E) = INS_LSR : AddrMode(&H4E) = ADDRESS_ABSOLUTE
        Ticks(&H4F) = 6 : Instruction(&H4F) = INS_NOP : AddrMode(&H4F) = ADDRESS_IMPLIED

        Ticks(&H50) = 2 : Instruction(&H50) = INS_BVC : AddrMode(&H50) = ADDRESS_RELATIVE
        Ticks(&H51) = 5 : Instruction(&H51) = INS_EOR : AddrMode(&H51) = ADDRESS_INDIRECT_Y
        Ticks(&H52) = 2 : Instruction(&H52) = INS_NOP : AddrMode(&H52) = ADDRESS_IMPLIED
        Ticks(&H53) = 8 : Instruction(&H53) = INS_NOP : AddrMode(&H53) = ADDRESS_IMPLIED
        Ticks(&H54) = 4 : Instruction(&H54) = INS_NOP : AddrMode(&H54) = ADDRESS_IMPLIED
        Ticks(&H55) = 4 : Instruction(&H55) = INS_EOR : AddrMode(&H55) = ADDRESS_ZERO_PAGE_X
        Ticks(&H56) = 6 : Instruction(&H56) = INS_LSR : AddrMode(&H56) = ADDRESS_ZERO_PAGE_X
        Ticks(&H57) = 6 : Instruction(&H57) = INS_NOP : AddrMode(&H57) = ADDRESS_IMPLIED
        Ticks(&H58) = 2 : Instruction(&H58) = INS_CLI : AddrMode(&H58) = ADDRESS_IMPLIED
        Ticks(&H59) = 4 : Instruction(&H59) = INS_EOR : AddrMode(&H59) = ADDRESS_ABSOLUTE_Y
        Ticks(&H5A) = 2 : Instruction(&H5A) = INS_NOP : AddrMode(&H5A) = ADDRESS_IMPLIED
        Ticks(&H5B) = 7 : Instruction(&H5B) = INS_NOP : AddrMode(&H5B) = ADDRESS_IMPLIED
        Ticks(&H5C) = 4 : Instruction(&H5C) = INS_NOP : AddrMode(&H5C) = ADDRESS_IMPLIED
        Ticks(&H5D) = 4 : Instruction(&H5D) = INS_EOR : AddrMode(&H5D) = ADDRESS_ABSOLUTE_X
        Ticks(&H5E) = 7 : Instruction(&H5E) = INS_LSR : AddrMode(&H5E) = ADDRESS_ABSOLUTE_X
        Ticks(&H5F) = 7 : Instruction(&H5F) = INS_NOP : AddrMode(&H5F) = ADDRESS_IMPLIED

        Ticks(&H60) = 6 : Instruction(&H60) = INS_RTS : AddrMode(&H60) = ADDRESS_IMPLIED
        Ticks(&H61) = 6 : Instruction(&H61) = INS_ADC : AddrMode(&H61) = ADDRESS_INDIRECT_X
        Ticks(&H62) = 2 : Instruction(&H62) = INS_NOP : AddrMode(&H62) = ADDRESS_IMPLIED
        Ticks(&H63) = 8 : Instruction(&H63) = INS_NOP : AddrMode(&H63) = ADDRESS_IMPLIED
        Ticks(&H64) = 3 : Instruction(&H64) = INS_NOP : AddrMode(&H64) = ADDRESS_ZERO_PAGE
        Ticks(&H65) = 3 : Instruction(&H65) = INS_ADC : AddrMode(&H65) = ADDRESS_ZERO_PAGE
        Ticks(&H66) = 5 : Instruction(&H66) = INS_ROR : AddrMode(&H66) = ADDRESS_ZERO_PAGE
        Ticks(&H67) = 5 : Instruction(&H67) = INS_NOP : AddrMode(&H67) = ADDRESS_IMPLIED
        Ticks(&H68) = 4 : Instruction(&H68) = INS_PLA : AddrMode(&H68) = ADDRESS_IMPLIED
        Ticks(&H69) = 2 : Instruction(&H69) = INS_ADC : AddrMode(&H69) = ADDRESS_IMMEDIATE
        Ticks(&H6A) = 2 : Instruction(&H6A) = INS_ROR : AddrMode(&H6A) = ADDRESS_ACCUMULATOR
        Ticks(&H6B) = 2 : Instruction(&H6B) = INS_NOP : AddrMode(&H6B) = ADDRESS_IMPLIED
        Ticks(&H6C) = 5 : Instruction(&H6C) = INS_JMP : AddrMode(&H6C) = ADDRESS_INDIRECT
        Ticks(&H6D) = 4 : Instruction(&H6D) = INS_ADC : AddrMode(&H6D) = ADDRESS_ABSOLUTE
        Ticks(&H6E) = 6 : Instruction(&H6E) = INS_ROR : AddrMode(&H6E) = ADDRESS_ABSOLUTE
        Ticks(&H6F) = 6 : Instruction(&H6F) = INS_NOP : AddrMode(&H6F) = ADDRESS_IMPLIED

        Ticks(&H70) = 2 : Instruction(&H70) = INS_BVS : AddrMode(&H70) = ADDRESS_RELATIVE
        Ticks(&H71) = 5 : Instruction(&H71) = INS_ADC : AddrMode(&H71) = ADDRESS_INDIRECT_Y
        Ticks(&H72) = 2 : Instruction(&H72) = INS_NOP : AddrMode(&H72) = ADDRESS_IMPLIED
        Ticks(&H73) = 8 : Instruction(&H73) = INS_NOP : AddrMode(&H73) = ADDRESS_IMPLIED
        Ticks(&H74) = 4 : Instruction(&H74) = INS_NOP : AddrMode(&H74) = ADDRESS_IMPLIED
        Ticks(&H75) = 4 : Instruction(&H75) = INS_ADC : AddrMode(&H75) = ADDRESS_ZERO_PAGE_X
        Ticks(&H76) = 6 : Instruction(&H76) = INS_ROR : AddrMode(&H76) = ADDRESS_ZERO_PAGE_X
        Ticks(&H77) = 6 : Instruction(&H77) = INS_NOP : AddrMode(&H77) = ADDRESS_IMPLIED
        Ticks(&H78) = 2 : Instruction(&H78) = INS_SEI : AddrMode(&H78) = ADDRESS_IMPLIED
        Ticks(&H79) = 4 : Instruction(&H79) = INS_ADC : AddrMode(&H79) = ADDRESS_ABSOLUTE_Y
        Ticks(&H7A) = 2 : Instruction(&H7A) = INS_NOP : AddrMode(&H7A) = ADDRESS_IMPLIED
        Ticks(&H7B) = 7 : Instruction(&H7B) = INS_NOP : AddrMode(&H7B) = ADDRESS_IMPLIED
        Ticks(&H7C) = 4 : Instruction(&H7C) = INS_NOP : AddrMode(&H7C) = ADDRESS_IMPLIED
        Ticks(&H7D) = 4 : Instruction(&H7D) = INS_ADC : AddrMode(&H7D) = ADDRESS_ABSOLUTE_X
        Ticks(&H7E) = 7 : Instruction(&H7E) = INS_ROR : AddrMode(&H7E) = ADDRESS_ABSOLUTE_X
        Ticks(&H7F) = 7 : Instruction(&H7F) = INS_NOP : AddrMode(&H7F) = ADDRESS_IMPLIED

        Ticks(&H80) = 2 : Instruction(&H80) = INS_NOP : AddrMode(&H80) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H81) = 6 : Instruction(&H81) = INS_STA : AddrMode(&H81) = ADDRESS_INDIRECT_X
        Ticks(&H82) = 2 : Instruction(&H82) = INS_NOP : AddrMode(&H82) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H83) = 6 : Instruction(&H83) = INS_NOP : AddrMode(&H83) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H84) = 3 : Instruction(&H84) = INS_STY : AddrMode(&H84) = ADDRESS_ZERO_PAGE
        Ticks(&H85) = 3 : Instruction(&H85) = INS_STA : AddrMode(&H85) = ADDRESS_ZERO_PAGE
        Ticks(&H86) = 3 : Instruction(&H86) = INS_STX : AddrMode(&H86) = ADDRESS_ZERO_PAGE
        Ticks(&H87) = 3 : Instruction(&H87) = INS_NOP : AddrMode(&H87) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H88) = 2 : Instruction(&H88) = INS_DEY : AddrMode(&H88) = ADDRESS_IMPLIED
        Ticks(&H89) = 2 : Instruction(&H89) = INS_NOP : AddrMode(&H89) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H8A) = 2 : Instruction(&H8A) = INS_TXA : AddrMode(&H8A) = ADDRESS_IMPLIED
        Ticks(&H8B) = 2 : Instruction(&H8B) = INS_NOP : AddrMode(&H8B) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H8C) = 4 : Instruction(&H8C) = INS_STY : AddrMode(&H8C) = ADDRESS_ABSOLUTE
        Ticks(&H8D) = 4 : Instruction(&H8D) = INS_STA : AddrMode(&H8D) = ADDRESS_ABSOLUTE
        Ticks(&H8E) = 4 : Instruction(&H8E) = INS_STX : AddrMode(&H8E) = ADDRESS_ABSOLUTE
        Ticks(&H8F) = 4 : Instruction(&H8F) = INS_NOP : AddrMode(&H8F) = ADDRESS_IMPLIED     'CORRECT

        Ticks(&H90) = 2 : Instruction(&H90) = INS_BCC : AddrMode(&H90) = ADDRESS_RELATIVE
        Ticks(&H91) = 6 : Instruction(&H91) = INS_STA : AddrMode(&H91) = ADDRESS_INDIRECT_Y
        Ticks(&H92) = 2 : Instruction(&H92) = INS_NOP : AddrMode(&H92) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H93) = 6 : Instruction(&H93) = INS_NOP : AddrMode(&H93) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H94) = 4 : Instruction(&H94) = INS_STY : AddrMode(&H94) = ADDRESS_ZERO_PAGE_X
        Ticks(&H95) = 4 : Instruction(&H95) = INS_STA : AddrMode(&H95) = ADDRESS_ZERO_PAGE_X
        Ticks(&H96) = 4 : Instruction(&H96) = INS_STX : AddrMode(&H96) = ADDRESS_ZERO_PAGE_Y
        Ticks(&H97) = 4 : Instruction(&H97) = INS_NOP : AddrMode(&H97) = ADDRESS_IMPLIED
        Ticks(&H98) = 2 : Instruction(&H98) = INS_TYA : AddrMode(&H98) = ADDRESS_IMPLIED
        Ticks(&H99) = 5 : Instruction(&H99) = INS_STA : AddrMode(&H99) = ADDRESS_ABSOLUTE_Y
        Ticks(&H9A) = 2 : Instruction(&H9A) = INS_TXS : AddrMode(&H9A) = ADDRESS_IMPLIED
        Ticks(&H9B) = 5 : Instruction(&H9B) = INS_NOP : AddrMode(&H9B) = ADDRESS_IMPLIED
        Ticks(&H9C) = 5 : Instruction(&H9C) = INS_NOP : AddrMode(&H9C) = ADDRESS_IMPLIED
        Ticks(&H9D) = 5 : Instruction(&H9D) = INS_STA : AddrMode(&H9D) = ADDRESS_ABSOLUTE_X
        Ticks(&H9E) = 5 : Instruction(&H9E) = INS_NOP : AddrMode(&H9E) = ADDRESS_IMPLIED     'CORRECT
        Ticks(&H9F) = 5 : Instruction(&H9F) = INS_NOP : AddrMode(&H9F) = ADDRESS_IMPLIED     'CORRECT

        Ticks(&HA0) = 2 : Instruction(&HA0) = INS_LDY : AddrMode(&HA0) = ADDRESS_IMMEDIATE
        Ticks(&HA1) = 6 : Instruction(&HA1) = INS_LDA : AddrMode(&HA1) = ADDRESS_INDIRECT_X
        Ticks(&HA2) = 2 : Instruction(&HA2) = INS_LDX : AddrMode(&HA2) = ADDRESS_IMMEDIATE
        Ticks(&HA3) = 6 : Instruction(&HA3) = INS_NOP : AddrMode(&HA3) = ADDRESS_IMPLIED
        Ticks(&HA4) = 3 : Instruction(&HA4) = INS_LDY : AddrMode(&HA4) = ADDRESS_ZERO_PAGE
        Ticks(&HA5) = 3 : Instruction(&HA5) = INS_LDA : AddrMode(&HA5) = ADDRESS_ZERO_PAGE
        Ticks(&HA6) = 3 : Instruction(&HA6) = INS_LDX : AddrMode(&HA6) = ADDRESS_ZERO_PAGE
        Ticks(&HA7) = 3 : Instruction(&HA7) = INS_NOP : AddrMode(&HA7) = ADDRESS_IMPLIED
        Ticks(&HA8) = 2 : Instruction(&HA8) = INS_TAY : AddrMode(&HA8) = ADDRESS_IMPLIED
        Ticks(&HA9) = 2 : Instruction(&HA9) = INS_LDA : AddrMode(&HA9) = ADDRESS_IMMEDIATE
        Ticks(&HAA) = 2 : Instruction(&HAA) = INS_TAX : AddrMode(&HAA) = ADDRESS_IMPLIED
        Ticks(&HAB) = 2 : Instruction(&HAB) = INS_NOP : AddrMode(&HAB) = ADDRESS_IMPLIED
        Ticks(&HAC) = 4 : Instruction(&HAC) = INS_LDY : AddrMode(&HAC) = ADDRESS_ABSOLUTE
        Ticks(&HAD) = 4 : Instruction(&HAD) = INS_LDA : AddrMode(&HAD) = ADDRESS_ABSOLUTE
        Ticks(&HAE) = 4 : Instruction(&HAE) = INS_LDX : AddrMode(&HAE) = ADDRESS_ABSOLUTE
        Ticks(&HAF) = 4 : Instruction(&HAF) = INS_NOP : AddrMode(&HAF) = ADDRESS_IMPLIED

        Ticks(&HB0) = 2 : Instruction(&HB0) = INS_BCS : AddrMode(&HB0) = ADDRESS_RELATIVE
        Ticks(&HB1) = 5 : Instruction(&HB1) = INS_LDA : AddrMode(&HB1) = ADDRESS_INDIRECT_Y
        Ticks(&HB2) = 2 : Instruction(&HB2) = INS_NOP : AddrMode(&HB2) = ADDRESS_IMPLIED
        Ticks(&HB3) = 5 : Instruction(&HB3) = INS_NOP : AddrMode(&HB3) = ADDRESS_IMPLIED
        Ticks(&HB4) = 4 : Instruction(&HB4) = INS_LDY : AddrMode(&HB4) = ADDRESS_ZERO_PAGE_X
        Ticks(&HB5) = 4 : Instruction(&HB5) = INS_LDA : AddrMode(&HB5) = ADDRESS_ZERO_PAGE_X
        Ticks(&HB6) = 4 : Instruction(&HB6) = INS_LDX : AddrMode(&HB6) = ADDRESS_ZERO_PAGE_Y
        Ticks(&HB7) = 4 : Instruction(&HB7) = INS_NOP : AddrMode(&HB7) = ADDRESS_IMPLIED
        Ticks(&HB8) = 2 : Instruction(&HB8) = INS_CLV : AddrMode(&HB8) = ADDRESS_IMPLIED
        Ticks(&HB9) = 4 : Instruction(&HB9) = INS_LDA : AddrMode(&HB9) = ADDRESS_ABSOLUTE_Y
        Ticks(&HBA) = 2 : Instruction(&HBA) = INS_TSX : AddrMode(&HBA) = ADDRESS_IMPLIED
        Ticks(&HBB) = 4 : Instruction(&HBB) = INS_NOP : AddrMode(&HBB) = ADDRESS_IMPLIED
        Ticks(&HBC) = 4 : Instruction(&HBC) = INS_LDY : AddrMode(&HBC) = ADDRESS_ABSOLUTE_X
        Ticks(&HBD) = 4 : Instruction(&HBD) = INS_LDA : AddrMode(&HBD) = ADDRESS_ABSOLUTE_X
        Ticks(&HBE) = 4 : Instruction(&HBE) = INS_LDX : AddrMode(&HBE) = ADDRESS_ABSOLUTE_Y
        Ticks(&HBF) = 4 : Instruction(&HBF) = INS_NOP : AddrMode(&HBF) = ADDRESS_IMPLIED

        Ticks(&HC0) = 2 : Instruction(&HC0) = INS_CPY : AddrMode(&HC0) = ADDRESS_IMMEDIATE
        Ticks(&HC1) = 6 : Instruction(&HC1) = INS_CMP : AddrMode(&HC1) = ADDRESS_INDIRECT_X
        Ticks(&HC2) = 2 : Instruction(&HC2) = INS_NOP : AddrMode(&HC2) = ADDRESS_IMPLIED
        Ticks(&HC3) = 8 : Instruction(&HC3) = INS_NOP : AddrMode(&HC3) = ADDRESS_IMPLIED
        Ticks(&HC4) = 3 : Instruction(&HC4) = INS_CPY : AddrMode(&HC4) = ADDRESS_ZERO_PAGE
        Ticks(&HC5) = 3 : Instruction(&HC5) = INS_CMP : AddrMode(&HC5) = ADDRESS_ZERO_PAGE
        Ticks(&HC6) = 5 : Instruction(&HC6) = INS_DEC : AddrMode(&HC6) = ADDRESS_ZERO_PAGE
        Ticks(&HC7) = 5 : Instruction(&HC7) = INS_NOP : AddrMode(&HC7) = ADDRESS_IMPLIED
        Ticks(&HC8) = 2 : Instruction(&HC8) = INS_INY : AddrMode(&HC8) = ADDRESS_IMPLIED
        Ticks(&HC9) = 2 : Instruction(&HC9) = INS_CMP : AddrMode(&HC9) = ADDRESS_IMMEDIATE
        Ticks(&HCA) = 2 : Instruction(&HCA) = INS_DEX : AddrMode(&HCA) = ADDRESS_IMPLIED
        Ticks(&HCB) = 2 : Instruction(&HCB) = INS_NOP : AddrMode(&HCB) = ADDRESS_IMPLIED
        Ticks(&HCC) = 4 : Instruction(&HCC) = INS_CPY : AddrMode(&HCC) = ADDRESS_ABSOLUTE
        Ticks(&HCD) = 4 : Instruction(&HCD) = INS_CMP : AddrMode(&HCD) = ADDRESS_ABSOLUTE
        Ticks(&HCE) = 6 : Instruction(&HCE) = INS_DEC : AddrMode(&HCE) = ADDRESS_ABSOLUTE
        Ticks(&HCF) = 6 : Instruction(&HCF) = INS_NOP : AddrMode(&HCF) = ADDRESS_IMPLIED

        Ticks(&HD0) = 2 : Instruction(&HD0) = INS_BNE : AddrMode(&HD0) = ADDRESS_RELATIVE
        Ticks(&HD1) = 5 : Instruction(&HD1) = INS_CMP : AddrMode(&HD1) = ADDRESS_INDIRECT_Y
        Ticks(&HD2) = 2 : Instruction(&HD2) = INS_NOP : AddrMode(&HD2) = ADDRESS_IMPLIED
        Ticks(&HD3) = 8 : Instruction(&HD3) = INS_NOP : AddrMode(&HD3) = ADDRESS_IMPLIED
        Ticks(&HD4) = 4 : Instruction(&HD4) = INS_NOP : AddrMode(&HD4) = ADDRESS_IMPLIED
        Ticks(&HD5) = 4 : Instruction(&HD5) = INS_CMP : AddrMode(&HD5) = ADDRESS_ZERO_PAGE_X
        Ticks(&HD6) = 6 : Instruction(&HD6) = INS_DEC : AddrMode(&HD6) = ADDRESS_ZERO_PAGE_X
        Ticks(&HD7) = 6 : Instruction(&HD7) = INS_NOP : AddrMode(&HD7) = ADDRESS_IMPLIED
        Ticks(&HD8) = 2 : Instruction(&HD8) = INS_CLD : AddrMode(&HD8) = ADDRESS_IMPLIED
        Ticks(&HD9) = 4 : Instruction(&HD9) = INS_CMP : AddrMode(&HD9) = ADDRESS_ABSOLUTE_Y
        Ticks(&HDA) = 2 : Instruction(&HDA) = INS_NOP : AddrMode(&HDA) = ADDRESS_IMPLIED
        Ticks(&HDB) = 7 : Instruction(&HDB) = INS_NOP : AddrMode(&HDB) = ADDRESS_IMPLIED
        Ticks(&HDC) = 4 : Instruction(&HDC) = INS_NOP : AddrMode(&HDC) = ADDRESS_IMPLIED
        Ticks(&HDD) = 4 : Instruction(&HDD) = INS_CMP : AddrMode(&HDD) = ADDRESS_ABSOLUTE_X
        Ticks(&HDE) = 7 : Instruction(&HDE) = INS_DEC : AddrMode(&HDE) = ADDRESS_ABSOLUTE_X
        Ticks(&HDF) = 7 : Instruction(&HDF) = INS_NOP : AddrMode(&HDF) = ADDRESS_IMPLIED

        Ticks(&HE0) = 2 : Instruction(&HE0) = INS_CPX : AddrMode(&HE0) = ADDRESS_IMMEDIATE
        Ticks(&HE1) = 6 : Instruction(&HE1) = INS_SBC : AddrMode(&HE1) = ADDRESS_INDIRECT_X
        Ticks(&HE2) = 2 : Instruction(&HE2) = INS_NOP : AddrMode(&HE2) = ADDRESS_IMPLIED
        Ticks(&HE3) = 8 : Instruction(&HE3) = INS_NOP : AddrMode(&HE3) = ADDRESS_IMPLIED
        Ticks(&HE4) = 3 : Instruction(&HE4) = INS_CPX : AddrMode(&HE4) = ADDRESS_ZERO_PAGE
        Ticks(&HE5) = 3 : Instruction(&HE5) = INS_SBC : AddrMode(&HE5) = ADDRESS_ZERO_PAGE
        Ticks(&HE6) = 5 : Instruction(&HE6) = INS_INC : AddrMode(&HE6) = ADDRESS_ZERO_PAGE
        Ticks(&HE7) = 5 : Instruction(&HE7) = INS_NOP : AddrMode(&HE7) = ADDRESS_IMPLIED
        Ticks(&HE8) = 2 : Instruction(&HE8) = INS_INX : AddrMode(&HE8) = ADDRESS_IMPLIED
        Ticks(&HE9) = 2 : Instruction(&HE9) = INS_SBC : AddrMode(&HE9) = ADDRESS_IMMEDIATE
        Ticks(&HEA) = 2 : Instruction(&HEA) = INS_NOP : AddrMode(&HEA) = ADDRESS_IMPLIED
        Ticks(&HEB) = 2 : Instruction(&HEB) = INS_NOP : AddrMode(&HEB) = ADDRESS_IMPLIED
        Ticks(&HEC) = 4 : Instruction(&HEC) = INS_CPX : AddrMode(&HEC) = ADDRESS_ABSOLUTE
        Ticks(&HED) = 4 : Instruction(&HED) = INS_SBC : AddrMode(&HED) = ADDRESS_ABSOLUTE
        Ticks(&HEE) = 6 : Instruction(&HEE) = INS_INC : AddrMode(&HEE) = ADDRESS_ABSOLUTE
        Ticks(&HEF) = 6 : Instruction(&HEF) = INS_NOP : AddrMode(&HEF) = ADDRESS_IMPLIED

        Ticks(&HF0) = 2 : Instruction(&HF0) = INS_BEQ : AddrMode(&HF0) = ADDRESS_RELATIVE
        Ticks(&HF1) = 5 : Instruction(&HF1) = INS_SBC : AddrMode(&HF1) = ADDRESS_INDIRECT_Y
        Ticks(&HF2) = 2 : Instruction(&HF2) = INS_NOP : AddrMode(&HF2) = ADDRESS_IMPLIED
        Ticks(&HF3) = 8 : Instruction(&HF3) = INS_NOP : AddrMode(&HF3) = ADDRESS_IMPLIED
        Ticks(&HF4) = 4 : Instruction(&HF4) = INS_NOP : AddrMode(&HF4) = ADDRESS_IMPLIED
        Ticks(&HF5) = 4 : Instruction(&HF5) = INS_SBC : AddrMode(&HF5) = ADDRESS_ZERO_PAGE_X
        Ticks(&HF6) = 6 : Instruction(&HF6) = INS_INC : AddrMode(&HF6) = ADDRESS_ZERO_PAGE_X
        Ticks(&HF7) = 6 : Instruction(&HF7) = INS_NOP : AddrMode(&HF7) = ADDRESS_IMPLIED
        Ticks(&HF8) = 2 : Instruction(&HF8) = INS_SED : AddrMode(&HF8) = ADDRESS_IMPLIED
        Ticks(&HF9) = 4 : Instruction(&HF9) = INS_SBC : AddrMode(&HF9) = ADDRESS_ABSOLUTE_Y
        Ticks(&HFA) = 2 : Instruction(&HFA) = INS_NOP : AddrMode(&HFA) = ADDRESS_IMPLIED
        Ticks(&HFB) = 7 : Instruction(&HFB) = INS_NOP : AddrMode(&HFB) = ADDRESS_IMPLIED
        Ticks(&HFC) = 4 : Instruction(&HFC) = INS_NOP : AddrMode(&HFC) = ADDRESS_IMPLIED
        Ticks(&HFD) = 4 : Instruction(&HFD) = INS_SBC : AddrMode(&HFD) = ADDRESS_ABSOLUTE_X
        Ticks(&HFE) = 7 : Instruction(&HFE) = INS_INC : AddrMode(&HFE) = ADDRESS_ABSOLUTE_X
        Ticks(&HFF) = 7 : Instruction(&HFF) = INS_NOP : AddrMode(&HFF) = ADDRESS_IMPLIED
    End Sub
    Public Sub Reset_CPU()
        a = 0
        X = 0
        Y = 0
        P = &H20
        s = &HFF

        PC = ReadMemory(&HFFFC) + (ReadMemory(&HFFFD) * &H100)

        Debug.Print("Reset to $" & Hex$(PC) & "[" & PC & "]")
    End Sub
    Public Sub Execute_CPU()
        Dim f As Integer = CInt(Frames)

        While Frames = f
            Opcode = ReadMemory(PC)
            PC = PC + 1

            Tick_Count = Tick_Count + Ticks(Opcode)

            Select Case Instruction(Opcode)
                Case INS_JMP
                    AdrMode(Opcode)
                    PC = SavePC
                Case INS_LDA
                    AdrMode(Opcode)
                    a = ReadMemory(SavePC)
                    SetFlags(a)
                Case INS_LDX
                    AdrMode(Opcode)
                    X = ReadMemory(SavePC)
                    SetFlags(X)
                Case INS_LDY
                    AdrMode(Opcode)
                    Y = ReadMemory(SavePC)
                    SetFlags(Y)
                Case INS_BNE
                    If ((P And &H2) = 0) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                    Else
                        PC = PC + 1
                    End If
                Case INS_CMP
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)

                    If (a + &H100 - Value) > &HFF Then
                        P = P Or &H1
                    Else
                        P = (P And &HFE)
                    End If

                    Value = (a + &H100 - Value) And &HFF
                    SetFlags(Value)
                Case INS_STA
                    AdrMode(Opcode)
                    WriteMemory(SavePC, a)
                Case INS_BIT
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)

                    If (Value And a) Then
                        P = (P And &HFD)
                    Else
                        P = P Or &H2
                    End If
                    P = ((P And &H3F) Or (Value And &HC0))
                Case INS_BVC
                    If ((P And &H40) = 0) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_BEQ
                    If (P And &H2) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_INY
                    Y = (Y + 1) And &HFF
                    If (Y) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (Y And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_BPL
                    If ((P And &H80) = 0) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                    Else
                        PC = PC + 1
                    End If
                Case INS_DEX
                    X = (X - 1) And &HFF
                    If (X) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (X And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_INC
                    AdrMode(Opcode)
                    WriteMemory((SavePC), (ReadMemory(SavePC) + 1) And &HFF)
                    Value = ReadMemory(SavePC)
                    If (Value) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (Value And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_DEC
                    AdrMode(Opcode)
                    WriteMemory((SavePC), (ReadMemory(SavePC) - 1) And &HFF)
                    Value = ReadMemory(SavePC)
                    If (Value) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (Value And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_JSR
                    PC = PC + 1
                    WriteMemory(s + &H100&, (PC \ &H100&))
                    s = (s - 1) And &HFF
                    WriteMemory(s + &H100&, (PC And &HFF))
                    s = (s - 1) And &HFF
                    PC = PC - 1
                    AdrMode(Opcode)
                    PC = SavePC
                Case INS_AND
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)
                    a = (a And Value)
                    SetFlags(a)
                Case INS_NOP
                Case INS_BRK
                    PC = PC + 1
                    WriteMemory(&H100& + s, (PC \ &H100&) And &HFF)
                    s = (s - 1) And &HFF
                    WriteMemory(&H100& + s, (PC And &HFF))
                    s = (s - 1) And &HFF
                    WriteMemory(&H100& + s, P)
                    s = (s - 1) And &HFF
                    P = P Or &H14
                    PC = ReadMemory(&HFFFE&) + (ReadMemory(&HFFFF&) * &H100&)
                Case INS_ADC
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)

                    SaveFlags = (P And &H1)

                    Sum = a
                    Sum = (Sum + Value) And &HFF
                    Sum = (Sum + SaveFlags) And &HFF

                    If (Sum > &H7F) Or (Sum < -&H80) Then
                        P = P Or &H40
                    Else
                        P = (P And &HBF)
                    End If

                    Sum = a + (Value + SaveFlags)
                    If (Sum > &HFF) Then
                        P = P Or &H1
                    Else
                        P = (P And &HFE)
                    End If

                    a = Sum And &HFF
                    If (P And &H8) Then
                        P = (P And &HFE)
                        If ((a And &HF) > &H9) Then
                            a = (a + &H6) And &HFF
                        End If
                        If ((a And &HF0) > &H90) Then
                            a = (a + &H60) And &HFF
                            P = P Or &H1
                        End If
                    Else
                        Tick_Count = Tick_Count + 1
                    End If
                    SetFlags(a)
                Case INS_EOR
                    AdrMode(Opcode)
                    a = a Xor ReadMemory(SavePC)
                    If (a) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (a And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_ASL
                    If AddrMode(Opcode) = ADDRESS_ACCUMULATOR Then
                        P = (P And &HFE) Or ((a \ 128) And &H1)
                        a = (a * 2) And &HFF
                        SetFlags(a)
                    Else
                        AdrMode(Opcode)
                        Value = ReadMemory(SavePC)

                        P = (P And &HFE) Or ((Value \ 128) And &H1)
                        Value = (Value * 2) And &HFF

                        WriteMemory(SavePC, (Value And &HFF))
                        SetFlags(Value)
                    End If
                Case INS_BCC
                    If ((P And &H1) = 0) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_BCS
                    If (P And &H1) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_BMI
                    If (P And &H80) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_BVS
                    If (P And &H40) Then
                        AdrMode(Opcode)
                        PC = PC + SavePC
                        Tick_Count = Tick_Count + 1
                    Else
                        PC = PC + 1
                    End If
                Case INS_CLC : P = P And &HFE
                Case INS_CLD : P = P And &HF7
                Case INS_CLI : P = P And &HFB
                Case INS_CLV : P = P And &HBF
                Case INS_CPX
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)

                    If (X + &H100 - Value > &HFF) Then
                        P = P Or &H1
                    Else
                        P = (P And &HFE)
                    End If

                    Value = (X + &H100 - Value) And &HFF
                    SetFlags(Value)
                Case INS_CPY
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC)

                    If (Y + &H100 - Value > &HFF) Then
                        P = (P Or &H1)
                    Else
                        P = (P And &HFE)
                    End If
                    Value = (Y + &H100 - Value) And &HFF
                    SetFlags(Value)
                Case INS_DEY
                    Y = (Y - 1) And &HFF
                    If (Y) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (Y And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_INX
                    X = (X + 1) And &HFF
                    If (X) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (X And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_LSR
                    If AddrMode(Opcode) = ADDRESS_ACCUMULATOR Then
                        P = (P And &HFE) Or (a And &H1)
                        a = (a \ 2) And &HFF
                        If (a) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (a And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    Else
                        AdrMode(Opcode)
                        Value = ReadMemory(SavePC)

                        P = ((P And &HFE) Or (Value And &H1))

                        Value = (Value \ 2) And &HFF
                        WriteMemory(SavePC, (Value And &HFF))
                        If (Value) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (Value And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    End If
                Case INS_ORA
                    AdrMode(Opcode)
                    a = a Or ReadMemory(SavePC)
                    SetFlags(a)
                Case INS_PHA
                    WriteMemory(&H100& + s, a)
                    s = (s - 1) And &HFF
                Case INS_PHP
                    WriteMemory(&H100& + s, P)
                    s = (s - 1) And &HFF
                Case INS_PLA
                    s = (s + 1) And &HFF
                    a = ReadMemory(s + &H100)
                    If (a) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (a And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_PLP
                    s = (s + 1) And &HFF
                    P = ReadMemory(s + &H100) Or &H20
                Case INS_ROL
                    If AddrMode(Opcode) = ADDRESS_ACCUMULATOR Then
                        SaveFlags = (P And &H1)
                        P = (P And &HFE) Or ((a \ 128) And &H1)
                        a = (a * 2) And &HFF
                        a = a Or SaveFlags
                        If (a) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (a And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    Else
                        SaveFlags = (P And &H1)
                        AdrMode(Opcode)
                        Value = ReadMemory(SavePC)

                        P = (P And &HFE) Or ((Value \ 128) And &H1)

                        Value = (Value * 2) And &HFF
                        Value = Value Or SaveFlags

                        WriteMemory(SavePC, (Value And &HFF))

                        If (Value) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (Value And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    End If
                Case INS_ROR
                    If AddrMode(Opcode) = ADDRESS_ACCUMULATOR Then
                        SaveFlags = (P And &H1)
                        P = (P And &HFE) Or (a And &H1)
                        a = (a \ 2) And &HFF

                        If (SaveFlags) Then
                            a = a Or &H80
                        End If
                        If (a) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (a And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    Else
                        SaveFlags = (P And &H1)
                        AdrMode(Opcode)
                        Value = ReadMemory(SavePC)

                        P = (P And &HFE) Or (Value And &H1)
                        Value = (Value \ 2) And &HFF
                        If (SaveFlags) Then
                            Value = Value Or &H80
                        End If
                        WriteMemory(SavePC, Value And &HFF)
                        If (Value) Then
                            P = P And &HFD
                        Else
                            P = P Or &H2
                        End If
                        If (Value And &H80) Then
                            P = P Or &H80
                        Else
                            P = P And &H7F
                        End If
                    End If
                Case INS_RTI
                    s = (s + 1) And &HFF
                    P = ReadMemory(s + &H100&) Or &H20
                    s = (s + 1) And &HFF
                    PC = ReadMemory(s + &H100&)
                    s = (s + 1) And &HFF
                    PC = PC + (ReadMemory(s + &H100) * &H100&)
                Case INS_RTS
                    s = (s + 1) And &HFF
                    PC = ReadMemory(s + &H100)
                    s = (s + 1) And &HFF
                    PC = PC + (ReadMemory(s + &H100) * &H100&)
                    PC = PC + 1
                Case INS_SBC
                    AdrMode(Opcode)
                    Value = ReadMemory(SavePC) Xor &HFF

                    SaveFlags = (P And &H1)

                    Sum = a
                    Sum = (Sum + Value) And &HFF
                    Sum = (Sum + (SaveFlags * 16)) And &HFF

                    If ((Sum > &H7F) Or (Sum <= -&H80)) Then
                        P = P Or &H40
                    Else
                        P = P And &HBF
                    End If

                    Sum = a + (Value + SaveFlags)

                    If (Sum > &HFF) Then
                        P = P Or &H1
                    Else
                        P = P And &HFE
                    End If

                    a = Sum And &HFF
                    If (P And &H8) Then
                        a = (a - &H66) And &HFF
                        P = P And &HFE
                        If ((a And &HF) > &H9) Then
                            a = (a + &H6) And &HFF
                        End If
                        If ((a And &HF0) > &H90) Then
                            a = (a + &H60) And &HFF
                            P = P Or &H1
                        End If
                    Else
                        Tick_Count = Tick_Count + 1
                    End If

                    If (a) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If

                    If (a And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_SEC : P = P Or &H1
                Case INS_SED : P = P Or &H8
                Case INS_SEI : P = P Or &H4
                Case INS_STX
                    AdrMode(Opcode)
                    WriteMemory(SavePC, X)
                Case INS_STY
                    AdrMode(Opcode)
                    WriteMemory(SavePC, Y)
                Case INS_TAX
                    X = a
                    If (X) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (X And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_TAY
                    Y = a
                    If (Y) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (Y And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_TXA
                    a = X
                    If (a) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (a And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_TYA
                    a = Y
                    If (a) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (a And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case INS_TXS
                    s = X
                Case INS_TSX
                    X = s
                    If (X) Then
                        P = P And &HFD
                    Else
                        P = P Or &H2
                    End If
                    If (X And &H80) Then
                        P = P Or &H80
                    Else
                        P = P And &H7F
                    End If
                Case Else
                    MsgBox("Invalid Opcode - " & Hex(Opcode))
            End Select

            If APU.Enabled = True Then
                Play_Sound_Loop(Sound_Buffer)
            Else
                Pause_Sound(Sound_Buffer)
            End If

            If Tick_Count > MaxCycles Then
                ExecutedCycles = ExecutedCycles + MaxCycles
                Mapper_HBlank(CurrentLine, PPU_Control2)
                Render_ScanLine(CurrentLine)
                If CurrentLine >= 240 Then
                    If CurrentLine = 240 Then
                        If Render Then Blit(Video_Buffer, 256, 240, Zoom, InterpolMode)
                        Frames = Frames + 1
                        ReadInput()
                        If (PPU_Control1 And &H80) Then CPU_NMI()
                        Application.DoEvents()
                    End If
                    PPU_Status = &H80
                End If

                If CurrentLine = 258 Then PPU_Status = 0

                If CurrentLine = 262 Then
                    'Update Sound
                    If APU.Enabled Then Sound_Process_Through_Mixer(ExecutedCycles)
                    'Next Frame
                    CurrentLine = 0
                    PPU_Status = 0
                Else
                    CurrentLine = CurrentLine + 1
                End If
                Tick_Count = Tick_Count - MaxCycles
            End If
        End While
    End Sub
    Public Sub AdrMode(ByVal Opcode As Byte)
        Select Case AddrMode(Opcode)
            Case ADDRESS_ABSOLUTE : SavePC = ReadMemory(PC) + (ReadMemory(PC + 1) * &H100&) : PC = PC + 2
            Case ADDRESS_ABSOLUTE_X
                SavePC = ReadMemory(PC)
                SavePC = SavePC + (ReadMemory(PC + 1) * &H100&)
                PC = PC + 2
                If (Ticks(Opcode) = 4) Then
                    If ((SavePC \ &H100&) = ((SavePC + X) \ &H100&)) Then
                    Else
                        Tick_Count = Tick_Count + 1
                    End If
                End If
                SavePC = SavePC + X
            Case ADDRESS_ABSOLUTE_Y
                SavePC = ReadMemory(PC) + (ReadMemory(PC + 1) * &H100&)
                PC = PC + 2
                If (Ticks(Opcode) = 4) Then
                    If ((SavePC \ &H100&) = ((SavePC + Y) \ &H100&)) Then
                    Else
                        Tick_Count = Tick_Count + 1
                    End If
                End If
                SavePC = SavePC + Y
            Case ADDRESS_IMPLIED 'Nothing
            Case ADDRESS_IMMEDIATE : SavePC = PC : PC = PC + 1
            Case ADDRESS_INDIRECT
                Help = ReadMemory(PC) + (ReadMemory(PC + 1) * &H100&)
                SavePC = ReadMemory(Help) + (ReadMemory(Help + 1) * &H100&)
                PC = PC + 2
            Case ADDRESS_INDIRECT_X
                Value = ReadMemory(PC) And &HFF
                Value = (Value + X) And &HFF
                PC = PC + 1
                SavePC = ReadMemory(Value) + (ReadMemory(Value + 1) * &H100&)
            Case ADDRESS_INDIRECT_Y
                Value = ReadMemory(PC)
                PC = PC + 1

                SavePC = ReadMemory(Value) + (ReadMemory(Value + 1) * &H100&)
                If (Ticks(Opcode) = 5) Then
                    If Not ((SavePC \ &H100&) = ((SavePC + Y) \ &H100&)) Then
                        Tick_Count = Tick_Count + 1
                    End If
                End If
                SavePC = SavePC + Y
            Case ADDRESS_RELATIVE : SavePC = ReadMemory(PC) : PC = PC + 1 : If (SavePC And &H80) Then SavePC = SavePC - &H100&
            Case ADDRESS_ZERO_PAGE : SavePC = ReadMemory(PC) : SavePC = SavePC And &HFF : PC = PC + 1
            Case ADDRESS_ZERO_PAGE_X
                SavePC = ReadMemory(PC)
                SavePC = SavePC + X
                PC = PC + 1
                SavePC = SavePC And &HFF
            Case ADDRESS_ZERO_PAGE_Y
                SavePC = ReadMemory(PC)
                SavePC = SavePC + Y
                PC = PC + 1
        End Select
    End Sub
    Public Sub SetFlags(ByVal Value As Byte)
        If (Value) Then
            P = P And &HFD
        Else
            P = P Or &H2
        End If
        If (Value And &H80) Then
            P = P Or &H80
        Else
            P = P And &H7F
        End If
    End Sub
    Public Sub CPU_IRQ()
        If (P And &H4) = 0 Then
            WriteMemory(&H100 + s, (PC \ &H100))
            s = (s - 1) And &HFF
            WriteMemory(&H100 + s, (PC And &HFF))
            s = (s - 1) And &HFF
            WriteMemory(&H100 + s, P)
            s = (s - 1) And &HFF
            P = P Or &H4
            PC = ReadMemory(&HFFFE) + (ReadMemory(&HFFFF) * &H100)
            Tick_Count = Tick_Count + 7
        End If
    End Sub
    Public Sub CPU_NMI()
        WriteMemory((s + &H100), (PC \ &H100))
        s = (s - 1) And &HFF
        WriteMemory((s + &H100), (PC And &HFF))
        s = (s - 1) And &HFF
        WriteMemory((s + &H100), P)
        P = P Or &H4
        s = (s - 1) And &HFF
        PC = ReadMemory(&HFFFA) + (ReadMemory(&HFFFB) * &H100)
        Tick_Count = Tick_Count + 7
    End Sub
End Module
