﻿Option Explicit On
Module APU_Noise


    Public Const APU_MODE_32k As Long = &H7FFF&
    Public Const APU_MODE_93 As Long = 93

    Public Function Noise_Channel_Render_Sample() As Long
        On Error Resume Next

        Dim Bit0 As Long
        Dim Bit1 As Long
        Dim Bit6 As Long

        With Noise_Channel
            If .Length_Counter > 0 Then
                '.SampleCount = Within_Range(.SampleCount + 1, 0, LONG_RANGE) 'causes minor slowdown
                .SampleCount = .SampleCount + 1
                If .SampleCount > .RenderedWavelength Then
                    .SampleCount = .SampleCount - .RenderedWavelength
                    If .Noise_Random_Generator_Mode = True Then '93 bits long
                        'According to http://wiki.nesdev.com/w/index.php/APU_Noise

                        'The shift register is 15 bits wide, with bits numbered
                        '14 - 13 - 12 - 11 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1 - 0

                        'Step 1) Feedback is calculated as the exclusive-OR of bit 0 and one other bit: bit 6 if Mode flag is set, otherwise bit 1.
                        '        Note: Exclusive Or is Xor, not Or! Or is Inclusive.
                        Bit0 = .NoiseShiftData \ Power_Of_2(14) 'Right Shifts it to bit 0 by right shifting 14 bits (see shift register table above)
                        Bit6 = .NoiseShiftData \ Power_Of_2(8) 'Right Shifts it to bit 6 by right shifting 8 bits (see shift register table above)
                        .Noise_Feedback = Bit0 Xor Bit6
                    Else '32767 bits long
                        'According to http://wiki.nesdev.com/w/index.php/APU_Noise

                        'The shift register is 15 bits wide, with bits numbered
                        '14 - 13 - 12 - 11 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1 - 0

                        'Step 1) Feedback is calculated as the exclusive-OR of bit 0 and one other bit: bit 6 if Mode flag is set, otherwise bit 1.
                        '        Note: Exclusive Or is Xor, not Or! Or is Inclusive.
                        Bit0 = .NoiseShiftData \ Power_Of_2(14) 'Right Shifts it to bit 0 by right shifting 14 bits (see shift register table above)
                        Bit1 = .NoiseShiftData \ Power_Of_2(13) 'Right Shifts it to bit 1 by right shifting 13 bits (see shift register table above)
                        .Noise_Feedback = Bit0 Xor Bit1
                    End If

                    'Step 2) The shift register is shifted right by one bit.
                    .NoiseShiftData = .NoiseShiftData * 2
                    'Step 3) Bit 14, the leftmost bit, is "set" to the feedback calculated earlier.
                    '        Note: You set it using the Or operator, not the And operator like you would see in most emulators' source code
                    .Noise_Feedback = .Noise_Feedback Or BIT_14

                    'Then the Noise shift gets updated and prepared to be outputed for sound.
                    .NoiseShiftData = .NoiseShiftData Or (.Noise_Feedback And &H1&)
                End If

                If .Envelope_Decay_Disable Then
                    Noise_Channel_Render_Sample = ((.NoiseShiftData And &H1&) * &H40&) * .Volume
                    Exit Function
                Else
                    Noise_Channel_Render_Sample = ((.NoiseShiftData And &H1&) * &H40&) * .Envelope_Counter
                    Exit Function
                End If
            Else
                .Volume = 0
            End If
            Noise_Channel_Render_Sample = 0
            Exit Function
        End With
    End Function

    Public Sub Noise_Channel_Update_Envelope()
        With Noise_Channel
            '.Envelope_Counter = .Envelope_Counter - 1
            'If .Envelope_Counter < 0 Then .Envelope_Counter = 0
            'If Not .Envelope_Decay_Disable And .Envelope > 0 And .Envelope_Counter = 0 Then
            '    .Envelope = .Envelope - 1
            '    .Envelope_Counter = .Volume
            'End If
            'If .Length_Counter_Disable And .Envelope = 0 Then .Envelope = 15

            If .Envelope_Start_Flag = True Then
                .Envelope_Start_Flag = False
                .Envelope = .Volume + 1
                .Envelope_Counter = 15
            Else
                .Envelope = .Envelope - 1
            End If

            If .Envelope <= 0 Then
                .Envelope = .Volume + 1
                If .Envelope_Counter > 0 Then
                    .Envelope_Counter = .Envelope_Counter - 1
                ElseIf .Length_Counter_Disable And .Envelope_Counter <= 0 Then
                    .Envelope_Counter = 15
                End If
            End If
        End With
    End Sub

    Public Sub Noise_Channel_Update_Frequency()
        With Noise_Channel
            .Frequency = NES_NTSC_CPU_SPEED_MHZ / ((.Wavelength + 1))
            .RenderedWavelength = .SamplingRate / .Frequency
        End With
    End Sub

    Public Sub Noise_Channel_Update_Length_Counter()
        With Noise_Channel
            If .Length_Counter_Disable = False Then .Length_Counter = .Length_Counter - 1
            If .Length_Counter <= 0 Then .Length_Counter = 0
        End With
    End Sub

    Public Sub Noise_Channel_Write_Register_1(ByVal Value As Byte)
        '$400C

        '$400C   --lE.eeee   l = Length Counter Disable / Envelope Decay Looping Enable
        '                    E = Envelope Decay Disable
        '                    e = Volume / Envelope Decay Rate

        'apu->noise.volume = value & 0x0F;
        'apu->noise.env_delay = decay_lut[value & 0x0F];
        'apu->noise.fixed_envelope = (value & 0x10) ? TRUE : FALSE;
        'apu->noise.holdnote = (value & 0x20) ? TRUE : FALSE;


        With Noise_Channel
            .Volume = Value And &HF& 'Where (BIT_0 + BIT_1 + BIT_2 + BIT_3) = Hex &HF&, Dec 15
            .Envelope_Counter = Value And &HF& 'Where (BIT_0 + BIT_1 + BIT_2 + BIT_3) = Hex &HF&, Dec 15
            .Envelope_Decay_Disable = (Value And BIT_4) <> 0
            .Length_Counter_Disable = (Value And BIT_5) <> 0 'Hold Note
        End With
    End Sub

    Public Sub Noise_Channel_Write_Register_2(ByVal Value As Byte)
        '$400E

        '$400E   M---.PPPP  M = Random Generator Mode
        '                   P = Period Index
        With Noise_Channel
            .Wavelength = .NoiseWavelengths(Value And &HF&) 'Where (BIT_0 + BIT_1 + BIT_2 + BIT_3) = Hex &HF&, Dec 15
            Noise_Channel_Update_Frequency()
            .Noise_Random_Generator_Mode = (Value And BIT_7) <> 0
        End With
    End Sub

    Public Sub Noise_Channel_Write_Register_3(ByVal Value As Byte)
        '$400F

        '$400F   llll.l---   Length Counter Load and Envelope Restart
        With Noise_Channel
            If .Enabled = True Then .Length_Counter = APU.Length_Values((Value And &HF8&) \ 8) 'where (BIT_3 + BIT_4 + BIT_5 + BIT_6 + BIT_7) = Hex &HF8&, Dec 248
        End With
    End Sub



End Module
