﻿Option Explicit On
Module MMU
    'Registers
    Public Register_8000 As Byte
    Public Register_A000 As Byte
    Public Register_C000 As Byte
    Public Register_E000 As Byte
    'Banks
    Public Bank0(2047) As Byte 'RAM
    Public Bank6(8191) As Byte 'SaveRAM
    Public Bank8(8191) As Byte '*PRG-ROM (8-E)
    Public BankA(8191) As Byte
    Public BankC(8191) As Byte
    Public BankE(8191) As Byte
    'Input Read
    Private Controller1_Count, Controller2_Count As Integer
    Public Function ReadMemory(ByVal Address As Integer)
        Dim Tmp As Byte

        Select Case Address
            Case &H0 To &H1FFF
                Return Bank0(Address And &H7FF) ' NES ram 0-7ff mirrored at 800 1000 1800
            Case &H2000
                Return PPU_Latch
            Case &H2001
                Return PPU_Latch
            Case &H2002
                Dim Ret As Byte

                PPU_AddressIsHi = True
                Ret = (PPU_Latch And &H1F) Or PPU_Status
                If (Ret And &H80) Then PPU_Status = (PPU_Status And &H60)

                Return Ret
            Case &H2004
                Tmp = PPU_Latch
                PPU_Latch = SpriteRAM(SpriteAddress)
                SpriteAddress = (SpriteAddress + 1) And &HFF
                Return Tmp
            Case &H2005
                Return PPU_Latch
            Case &H2006
                Return PPU_Latch
            Case &H2007
                Dim VAddr As Integer

                Tmp = PPU_Latch
                VAddr = (Loopy_V And &H3FFF)
                If (PPU_Control1 And &H4) Then
                    Loopy_V = Loopy_V + 32
                Else
                    Loopy_V = Loopy_V + 1
                End If

                If VAddr >= &H2000& And VAddr <= &H3EFF& Then '3000 - 3EFF Mirroring of 2000 - 2FFF
                    If VAddr >= &H3000 Then VAddr = VAddr And &HEFFF 'Handle Mirroring
                    PPU_Latch = Name_Table_Array(Mirror((VAddr And &HC00) \ &H400), VAddr And &H3FF)
                Else
                    '3F00-3F0F = BG Palette
                    '3F10-3F1F = SPR Palette
                    If VAddr >= &H3F00 Then 'Palette
                        Tmp = Tmp And &H3F
                        If (VAddr And &HF) = 0 Then
                            PPU_Latch = VRAM(&H3F00)  'BG 0 Pal
                            PPU_Latch = VRAM(&H3F10)  'SPR 0 Pal
                        ElseIf (VAddr And &H10) = 0 Then
                            PPU_Latch = VRAM(&H3F00 + (VAddr And &HF))  'BG Pal
                        Else
                            PPU_Latch = VRAM(&H3F10 + (VAddr And &HF)) 'SPR Pal
                        End If
                    End If
                    PPU_Latch = VRAM(VAddr)
                End If

                Return Tmp
            Case &H4016
                Tmp = Controller_1(Controller1_Count)
                Controller1_Count = (Controller1_Count + 1) And 7
                Return Tmp
            Case &H4017
                Tmp = Controller_2(Controller2_Count)
                Controller2_Count = (Controller2_Count + 1) And 7
                Return Tmp
            Case &H6000 To &H7FFF
                Return Bank6(Address And &H1FFF)
            Case &H8000 To &H9FFF
                Return Bank8(Address And &H1FFF)
            Case &HA000 To &HBFFF
                Return BankA(Address And &H1FFF)
            Case &HC000 To &HDFFF
                Return BankC(Address And &H1FFF)
            Case &HE000 To &HFFFF
                Return BankE(Address And &H1FFF)
            Case Else
                Return Nothing
        End Select
    End Function
    Public Sub WriteMemory(ByVal Address As Integer, ByVal Value As Byte)
        On Error Resume Next
        If Address >= &H2000& And Address <= &H3FFF& Then PPU_Latch = Value
        Select Case Address
            Case &H0 To &H1FFF : Bank0(Address And &H7FF) = Value
            Case &H2000
                Loopy_T = (Loopy_T And &HF3FF) Or (Value And 3) * &H400

                If (Value And &H80) And Not (PPU_Control1 And &H80) And ((PPU_Latch And &H1F) And &H80) Then
                    CPU_NMI() 'hmm..
                End If

                PPU_Control1 = Value
            Case &H2001
                PPU_Control2 = Value
            Case &H2002
                PPU_Latch = Value
            Case &H2003
                SpriteAddress = Value
            Case &H2004
                SpriteRAM(SpriteAddress) = Value
                SpriteAddress = (SpriteAddress + 1) And &HFF
            Case &H2005
                If PPU_AddressIsHi Then 'Horizontal Scrolling
                    Loopy_T = (Loopy_T And &HFFE0&) Or Value \ 8
                    Loopy_X = Value And 7
                Else 'Vertical Scrolling
                    Loopy_T = (Loopy_T And &HFC1F&) Or ((Value And &HF8) * 4)
                    Loopy_T = (Loopy_T And &H8FFF&) Or ((Value And 7) * &H1000)
                End If
                PPU_AddressIsHi = Not PPU_AddressIsHi
            Case &H2006
                If PPU_AddressIsHi Then
                    Loopy_T = (Loopy_T And &HFF) Or ((Value And &H3F) * &H100&)
                Else
                    Loopy_T = (Loopy_T And &HFF00&) Or Value
                    Loopy_V = Loopy_T
                End If
                PPU_AddressIsHi = Not PPU_AddressIsHi
            Case &H2007
                Dim VAddr As Integer

                PPU_Latch = Value
                VAddr = (Loopy_V And &H3FFF&)
                If (PPU_Control1 And &H4) Then
                    Loopy_V = Loopy_V + 32
                Else
                    Loopy_V = Loopy_V + 1
                End If

                If VAddr >= &H2000& And VAddr <= &H3EFF& Then '3000 - 3EFF Mirroring of 2000 - 2FFF
                    If VAddr >= &H3000 Then VAddr = VAddr And &HEFFF& 'Handle Mirroring
                    Name_Table_Array(Mirror((VAddr And &HC00&) \ &H400&), VAddr And &H3FF&) = Value
                Else
                    '3F00-3F0F = BG Palette
                    '3F10-3F1F = SPR Palette
                    If VAddr >= &H3F00& Then 'Palette
                        Value = Value And &H3F
                        If (VAddr And &HF) = 0 Then
                            VRAM(&H3F00&) = Value 'BG 0 Pal
                            VRAM(&H3F10&) = Value 'SPR 0 Pal
                        ElseIf (VAddr And &H10) = 0 Then
                            VRAM(&H3F00& + (VAddr And &HF)) = Value 'BG Pal
                        Else
                            VRAM(&H3F10& + (VAddr And &HF)) = Value 'SPR Pal
                        End If
                    End If
                    VRAM(VAddr) = Value
                End If

                PPU_AddressIsHi = True
            Case &H4000 To &H4013
                WriteSound(Address, Value)
            Case &H4014
                Buffer.BlockCopy(Bank0, Value * &H100, SpriteRAM, 0, &H100)
            Case &H4015
                Sound_Channel_Write_Status_Register(Value)
            Case &H6000 To &H7FFF
                Bank6(Address And &H1FFF&) = Value
            Case &H8000& To &HFFFF& : Mapper_Write(Address, Value)
        End Select
    End Sub
    Public Sub SetupBanks()
        Register_8000 = MaskBankAddress(Register_8000)
        Register_A000 = MaskBankAddress(Register_A000)
        Register_C000 = MaskBankAddress(Register_C000)
        Register_E000 = MaskBankAddress(Register_E000)

        Buffer.BlockCopy(PROM, Register_8000 * &H2000, Bank8, 0, &H2000)
        Buffer.BlockCopy(PROM, Register_A000 * &H2000, BankA, 0, &H2000)
        Buffer.BlockCopy(PROM, Register_C000 * &H2000, BankC, 0, &H2000)
        Buffer.BlockCopy(PROM, Register_E000 * &H2000, BankE, 0, &H2000)
    End Sub
    Private Sub CopyBanks(ByVal Dest, ByVal Src, ByVal Count)
        On Error Resume Next
        Buffer.BlockCopy(VROM, Src * &H400, VRAM, Dest * &H400, Count * &H400)
    End Sub
    Public Function MaskBankAddress(ByVal Bank As Byte)
        If Bank >= PrgCount * 2 Then
            Dim i As Byte = &HFF
            Do While (Bank And i) >= PrgCount * 2
                i = i \ 2
            Loop
            MaskBankAddress = (Bank And i)
        Else
            MaskBankAddress = Bank
        End If
    End Function
    Public Function MaskVROM(ByVal Page As Byte, ByVal Mask As Long) As Byte
        Dim i As Long
        If Mask = 0 Then Mask = 256
        If Mask And Mask - 1 Then
            i = 1
            Do While i < Mask
                i = i + i
            Loop
        Else
            i = Mask
        End If
        i = (Page And (i - 1))
        If i >= Mask Then i = Mask - 1
        MaskVROM = i
    End Function
    Public Sub Select8KVROM(ByVal Val1 As Byte)
        Val1 = MaskVROM(Val1, ChrCount)
        CopyBanks(0, Val1 * 8, 8)
    End Sub
    Public Sub Select4KVROM(ByVal Val1 As Byte, ByVal Bank As Byte)
        Val1 = MaskVROM(Val1, ChrCount * 2)
        CopyBanks(Bank * 4, Val1 * 4, 4)
    End Sub
    Public Sub Select2KVROM(ByVal Val1 As Byte, ByVal Bank As Byte)
        Val1 = MaskVROM(Val1, ChrCount * 4)
        CopyBanks(Bank * 2, Val1 * 2, 2)
    End Sub
    Public Sub Select1KVROM(ByVal Val1 As Byte, ByVal Bank As Byte)
        Val1 = MaskVROM(Val1, ChrCount * 8)
        CopyBanks(Bank, Val1, 1)
    End Sub
    Public Sub Select8KPROM(ByVal Val1 As Byte, ByVal Bank As Byte)
        If Bank > 3 Then Bank -= 4
        Select Case Bank
            Case 0
                Register_8000 = Val1
            Case 1
                Register_A000 = Val1
            Case 2
                Register_C000 = Val1
            Case 3
                Register_E000 = Val1
        End Select
        SetupBanks()
    End Sub
    Public Sub Select16KPROM(ByVal Val1 As Byte, ByVal Bank As Byte)
        If Bank > 3 Then Bank -= 4
        Select Case Bank
            Case 0
                Register_8000 = Val1
                Register_A000 = Val1 + 1
            Case 1
                Register_A000 = Val1
                Register_C000 = Val1 + 1
            Case 2
                Register_C000 = Val1
                Register_E000 = Val1 + 1
        End Select
        SetupBanks()
    End Sub
    Public Sub Select16KPROM(ByVal Val1 As Byte, ByVal Val2 As Byte, ByVal Bank As Byte)
        If Bank > 3 Then Bank -= 4
        Select Case Bank
            Case 0
                Register_8000 = Val1
                Register_A000 = Val2
            Case 1
                Register_A000 = Val1
                Register_C000 = Val2
            Case 2
                Register_C000 = Val1
                Register_E000 = Val2
        End Select
        SetupBanks()
    End Sub
    Public Sub Select32KPROM(ByVal Val1 As Byte)
        Register_8000 = Val1
        Register_A000 = Val1 + 1
        Register_C000 = Val1 + 2
        Register_E000 = Val1 + 3
        SetupBanks()
    End Sub
    Public Sub Select32KPROM(ByVal Val1 As Byte, ByVal Val2 As Byte, ByVal Val3 As Byte, ByVal Val4 As Byte)
        Register_8000 = Val1
        Register_A000 = Val2
        Register_C000 = Val3
        Register_E000 = Val4
        SetupBanks()
    End Sub
End Module
