﻿Option Explicit On
Imports System.IO
Module Main
    'ROM Arrays
    Public PROM() As Byte
    Public VROM() As Byte
    'Header info
    Public PrgCount As Byte, ChrCount As Byte
    Public Mapper As Byte, Mirroring As Byte, Trainer As Byte, FourScreen As Byte
    Public UsesSRAM As Boolean 'SaveRAM
    'Misc
    Public NES_On, NES_Pause As Boolean
    Public Frames As Long
    'Input
    Public KeyCodes(255) As Byte
    Public Controller_1(7) As Byte
    Public Controller_2(7) As Byte
    'LookUp Tables
    Public Power_Of_2(31) As Long
    Public TLook(65536 * 8 - 1) As Byte
    Public Sub LoadROM(ByVal FileName As String)
        DestroyNes()

        Dim Input As New FileStream(FileName, FileMode.Open)
        Dim ROM(CInt(Input.Length - 1)) As Byte
        Input.Read(ROM, 0, CInt(Input.Length))

        ' Parse Header
        '---------------------------------------
        Dim PrgMark As Integer
        Dim ROMCtrl As Byte
        Dim ROMCtrl2 As Byte

        PrgCount = ROM(4)
        ChrCount = ROM(5)
        ROMCtrl = ROM(6)
        ROMCtrl2 = ROM(7)

        Mapper = (ROMCtrl And &HF0) \ 16
        Mapper = Mapper + ROMCtrl2

        Mirroring = ROMCtrl And &H1
        Trainer = ROMCtrl And &H4
        FourScreen = ROMCtrl And &H8
        If ROMCtrl And &H2 Then UsesSRAM = True
        PrgMark = (PrgCount * &H4000&) - 1

        ' Fill PRG-ROM and Video ROM
        '---------------------------------------
        ReDim PROM(PrgMark)
        Dim StartAt As Integer
        If Trainer Then StartAt = 528 Else StartAt = 16
        Buffer.BlockCopy(ROM, StartAt, PROM, 0, UBound(PROM))

        ReDim VROM(ChrCount * &H2000&)
        PrgMark = &H4000 * PrgCount + StartAt
        If ChrCount Then Buffer.BlockCopy(ROM, PrgMark, VROM, 0, UBound(VROM))

        ' Reset NES
        '---------------------------------------
        Mapper_Reset()
        Reset_CPU()
        For i As Integer = 0 To 7 'Erase Input
            Controller_1(i) = &H40
            Controller_2(i) = &H40
        Next

        ' Setup Mirroring
        '---------------------------------------
        If Mirroring = 1 Then MirrorXor = &H800 Else MirrorXor = &H400
        Do_Mirroring()

        ' Load NES Palette
        '---------------------------------------
        LoadDefaultPalette()

        Run() 'Turn on the NES!
    End Sub
    Public Sub Run()
        NES_On = True
        Do While NES_On
            Do While NES_Pause
                Application.DoEvents()
            Loop
            Execute_CPU()
        Loop
    End Sub
    Public Sub ReadInput()
        Controller_1(0) = KeyCodes(Keys.X)
        Controller_1(1) = KeyCodes(Keys.Z)
        Controller_1(2) = KeyCodes(Keys.Tab)
        Controller_1(3) = KeyCodes(Keys.Enter)
        Controller_1(4) = KeyCodes(Keys.Up)
        Controller_1(5) = KeyCodes(Keys.Down)
        Controller_1(6) = KeyCodes(Keys.Left)
        Controller_1(7) = KeyCodes(Keys.Right)
    End Sub
    Public Sub LoadDefaultPalette()
        Dim Current_Color As Long

        'Row 1
        Palette(0) = NES_RGB(104, 104, 104)
        Palette(1) = NES_RGB(0, 42, 136)
        Palette(2) = NES_RGB(20, 18, 167)
        Palette(3) = NES_RGB(59, 0, 164)
        Palette(4) = NES_RGB(92, 0, 126)
        Palette(5) = NES_RGB(110, 0, 64)
        Palette(6) = NES_RGB(108, 6, 0)
        Palette(7) = NES_RGB(86, 29, 0)
        Palette(8) = NES_RGB(51, 53, 0)
        Palette(9) = NES_RGB(11, 72, 0)
        Palette(10) = NES_RGB(0, 82, 0)
        Palette(11) = NES_RGB(0, 79, 8)
        Palette(12) = NES_RGB(0, 64, 77)
        Palette(13) = NES_RGB(0, 0, 0)
        Palette(14) = NES_RGB(0, 0, 0)
        Palette(15) = NES_RGB(0, 0, 0)

        'Row 2
        Palette(16) = NES_RGB(173, 173, 173)
        Palette(17) = NES_RGB(21, 95, 217)
        Palette(18) = NES_RGB(66, 64, 255)
        Palette(19) = NES_RGB(117, 39, 254)
        Palette(20) = NES_RGB(160, 26, 204)
        Palette(21) = NES_RGB(183, 30, 123)
        Palette(22) = NES_RGB(141, 49, 12)
        Palette(23) = NES_RGB(113, 63, 0)
        Palette(24) = NES_RGB(100, 99, 46)
        Palette(25) = NES_RGB(56, 135, 0)
        Palette(26) = NES_RGB(0, 110, 0)
        Palette(27) = NES_RGB(0, 143, 50)
        Palette(28) = NES_RGB(0, 104, 101)
        Palette(29) = NES_RGB(0, 0, 0)
        Palette(30) = NES_RGB(0, 0, 0)
        Palette(31) = NES_RGB(0, 0, 0)

        'Row 3
        Palette(32) = NES_RGB(255, 254, 255)
        Palette(33) = NES_RGB(100, 176, 255)
        Palette(34) = NES_RGB(146, 154, 255)
        Palette(35) = NES_RGB(198, 118, 255)
        Palette(36) = NES_RGB(243, 106, 255)
        Palette(37) = NES_RGB(254, 110, 204)
        Palette(38) = NES_RGB(254, 129, 112)
        Palette(39) = NES_RGB(225, 184, 114)
        Palette(40) = NES_RGB(188, 190, 0)
        Palette(41) = NES_RGB(175, 196, 0)
        Palette(42) = NES_RGB(92, 228, 48)
        Palette(43) = NES_RGB(69, 224, 130)
        Palette(44) = NES_RGB(72, 205, 222)
        Palette(45) = NES_RGB(79, 79, 79)
        Palette(46) = NES_RGB(0, 0, 0)
        Palette(47) = NES_RGB(0, 0, 0)

        'Row 4
        Palette(48) = NES_RGB(255, 254, 255)
        Palette(49) = NES_RGB(192, 223, 255)
        Palette(50) = NES_RGB(211, 210, 255)
        Palette(51) = NES_RGB(232, 200, 255)
        Palette(52) = NES_RGB(251, 194, 255)
        Palette(53) = NES_RGB(254, 196, 234)
        Palette(54) = NES_RGB(254, 204, 197)
        Palette(55) = NES_RGB(247, 236, 225)
        Palette(56) = NES_RGB(228, 229, 148)
        Palette(57) = NES_RGB(207, 239, 150)
        Palette(58) = NES_RGB(189, 244, 171)
        Palette(59) = NES_RGB(179, 243, 204)
        Palette(60) = NES_RGB(181, 235, 242)
        Palette(61) = NES_RGB(184, 184, 184)
        Palette(62) = NES_RGB(0, 0, 0)
        Palette(63) = NES_RGB(0, 0, 0)

        For Current_Color = 0 To 63
            Palette(Current_Color + 64) = Palette(Current_Color)
            Palette(Current_Color + 128) = Palette(Current_Color)
            Palette(Current_Color + 192) = Palette(Current_Color)
        Next Current_Color
    End Sub
    Function NES_RGB(ByVal B As Integer, ByVal G As Integer, ByVal R As Integer)
        Return B * 65536 + G * 256 + R
    End Function
    Public Sub FillTLook()
        Dim b1, b2, c, X
        For b1 = 0 To 255
            For b2 = 0 To 255
                For X = 0 To 7
                    If b1 And Power_Of_2(X) Then c = 1 Else c = 0
                    If b2 And Power_Of_2(X) Then c = c + 2
                    TLook(b1 * 2048 + b2 * 8 + X) = c
                Next X
        Next b2, b1
    End Sub
    Public Sub DestroyNes()
        'Turn off the NES
        NES_On = False
        'Erase Video Buffer
        Array.Clear(Video_Buffer, 0, Video_Buffer.Length)
        'Erase ROM Data
        Erase PROM
        Erase VROM
        'Erase RAM Stuff
        Array.Clear(VRAM, 0, VRAM.Length)
        Array.Clear(Bank0, 0, Bank0.Length)
        Array.Clear(Bank6, 0, Bank6.Length)
        Array.Clear(Bank8, 0, Bank8.Length)
        Array.Clear(BankA, 0, BankA.Length)
        Array.Clear(BankC, 0, BankC.Length)
        Array.Clear(BankE, 0, BankE.Length)
    End Sub
    Public Sub Fill_Array(ByVal a() As Long, ByVal ParamArray B() As Object)
        For i As Integer = 0 To UBound(a)
            a(i) = B(i)
        Next i
    End Sub
    Public Sub Fill_Array_Byte(ByVal a() As Byte, ByVal ParamArray B() As Object)
        For i As Integer = 0 To UBound(a)
            a(i) = B(i)
        Next i
    End Sub
    Public ReadOnly Property Timer As Double
        Get
            Return (CDbl((DateTime.Now.Ticks Mod &HC92A69C000)) / 10000000)
        End Get
    End Property
End Module
