﻿Option Explicit On
Module Mappers
    'MMC1
    Private MMC1_Accumulator, MMC1_Sequence As Integer
    Private MMC1_Register() As Byte
    'MMC3
    Private MMC3_Command As Byte
    Private MMC3_PrgAddr As Byte
    Private MMC3_ChrAddr As Integer
    Private MMC3_IrqVal As Byte
    Private MMC3_TmpVal As Byte
    Private MMC3_IrqOn As Boolean
    Private MMC3_Swap As Boolean
    Public PrgSwitch1 As Byte
    Public PrgSwitch2 As Byte
    Public Sub Mapper_Reset()
        Select Case Mapper
            Case 0
                DefaultReset()
            Case 1
                Mapper1_Reset()
            Case 2
                Mapper2_Reset()
            Case 4
                Mapper4_Reset()
            Case Else
                MsgBox("Mapper " & Mapper & " is not supported yet!", vbCritical, "Error!")
        End Select
    End Sub
    Public Sub Mapper_Write(ByVal Address As Integer, ByVal Value As Byte)
        Select Case Mapper
            Case 0
                'Nothing to do here!
            Case 1
                Mapper1_Write(Address, Value)
            Case 2
                Mapper2_Write(Address, Value)
            Case 4
                Mapper4_Write(Address, Value)
        End Select
    End Sub
    Public Sub Mapper_HBlank(ByVal ScanLine As Integer, Optional ByVal Two As Byte = 0)
        Select Case Mapper
            Case 4
                Mapper4_HBlank(ScanLine, Two)
        End Select
    End Sub
    Private Sub DefaultReset()
        Select32KPROM(0, 1, &HFE, &HFF)
        If ChrCount Then Select8KVROM(0)
    End Sub
    '+==========+
    '| Mapper 1 |
    '+==========+
    Private Sub Mapper1_Reset()
        Erase MMC1_Register
        ReDim MMC1_Register(4)
        MMC1_Sequence = 0
        MMC1_Accumulator = 0
        DefaultReset()
    End Sub
    Private Sub Mapper1_Write(ByVal Address As Integer, ByVal Value As Byte)
        Dim Bank_Select As Integer

        If Value And &H80 Then
            MMC1_Register(0) = MMC1_Register(0) Or &HC
            MMC1_Accumulator = MMC1_Register((Address \ &H2000) And 3)
        Else
            If Value And 1 Then
                MMC1_Accumulator = MMC1_Accumulator Or Power_Of_2(MMC1_Sequence)
            End If
            MMC1_Sequence += 1
        End If

        If MMC1_Sequence = 5 Then
            MMC1_Register(Address \ &H2000& And 3) = MMC1_Accumulator
            MMC1_Sequence = 0
            MMC1_Accumulator = 0

            If (PrgCount = &H20) Then '/* 512k cart */'
                Bank_Select = (MMC1_Register(1) And &H10) * 2
            Else '/* other carts */'
                Bank_Select = 0
            End If

            If MMC1_Register(0) And 2 Then 'enable panning
                Mirroring = (MMC1_Register(0) And 1) Xor 1
            Else 'disable panning
                Mirroring = 2
            End If
            Do_Mirroring()

            If (MMC1_Register(0) And 8) = 0 Then
                Select32KPROM(4 * (MMC1_Register(3) And 15) + Bank_Select)
            ElseIf (MMC1_Register(0) And 4) Then '16k
                Select16KPROM(&HFE, 6)
                Select16KPROM(((MMC1_Register(3) And 15) * 2) + Bank_Select, 4)
            Else '32k
                Select16KPROM(0, 4)
                Select16KPROM(((MMC1_Register(3) And 15) * 2) + Bank_Select, 6)
            End If

            If (MMC1_Register(0) And &H10) Then '4k
                Select4KVROM(MMC1_Register(1), 0)
                Select4KVROM(MMC1_Register(2), 1)
            Else '8k
                Select8KVROM(MMC1_Register(1) \ 2)
            End If
        End If
    End Sub
    '+==========+
    '| Mapper 2 |
    '+==========+
    Private Sub Mapper2_Reset()
        DefaultReset()
    End Sub
    Private Sub Mapper2_Write(ByVal Address As Integer, ByVal Value As Byte)
        Select16KPROM(Value * 2, 4)
    End Sub
    '+==========+
    '| Mapper 4 |
    '+==========+
    Private Sub Mapper4_Reset()
        MMC3_Command = 0
        MMC3_PrgAddr = 0
        MMC3_ChrAddr = 0
        MMC3_IrqVal = 0
        MMC3_TmpVal = 0
        MMC3_IrqOn = False
        MMC3_Swap = False
        PrgSwitch1 = 0
        PrgSwitch2 = 0
        DefaultReset()
        Mapper4_Sync()
    End Sub
    Public Sub Mapper4_Write(ByVal Address As Long, ByVal Value As Byte)
        Select Case Address
            Case &H8000
                MMC3_Command = Value And &H7
                If Value And &H80 Then MMC3_ChrAddr = &H1000& Else MMC3_ChrAddr = 0
                If Value And &H40 Then MMC3_Swap = True Else MMC3_Swap = False
            Case &H8001
                Select Case MMC3_Command
                    Case 0 : Select1KVROM(Value, 0) : Select1KVROM(Value + 1, 1)
                    Case 1 : Select1KVROM(Value, 2) : Select1KVROM(Value + 1, 3)
                    Case 2 : Select1KVROM(Value, 4)
                    Case 3 : Select1KVROM(Value, 5)
                    Case 4 : Select1KVROM(Value, 6)
                    Case 5 : Select1KVROM(Value, 7)
                    Case 6 : PrgSwitch1 = Value : Mapper4_Sync()
                    Case 7 : PrgSwitch2 = Value : Mapper4_Sync()
                End Select
            Case &HA000
                If (Value And &H1) Then Mirroring = 0 Else Mirroring = 1
                Do_Mirroring()
            Case &HA001 : If Value Then UsesSRAM = True Else UsesSRAM = False
            Case &HC000 : MMC3_IrqVal = Value
            Case &HC001 : MMC3_TmpVal = Value
            Case &HE000 : MMC3_IrqOn = False : MMC3_IrqVal = MMC3_TmpVal
            Case &HE001 : MMC3_IrqOn = True
        End Select
    End Sub
    Public Sub Mapper4_HBlank(ByVal ScanLine As Integer, ByVal Two As Byte)
        If ScanLine = 0 Then
            MMC3_IrqVal = MMC3_TmpVal
        ElseIf ScanLine > 239 Then
            Exit Sub
        ElseIf MMC3_IrqOn And (Two And &H18) Then
            MMC3_IrqVal = (MMC3_IrqVal - 1) And &HFF
            If (MMC3_IrqVal = 0) Then
                CPU_IRQ()
                MMC3_IrqVal = MMC3_TmpVal
            End If
        End If
    End Sub
    Public Sub Mapper4_Sync()
        If MMC3_Swap Then
            Select32KPROM(&HFE, PrgSwitch2, PrgSwitch1, &HFF)
        Else
            Select32KPROM(PrgSwitch1, PrgSwitch2, &HFE, &HFF)
        End If
    End Sub
End Module
