﻿Option Explicit On
Imports System.Drawing.Drawing2D
Module PPU
    Public Render As Boolean = True
    Public Zoom As Integer = 1
    Public InterpolMode As Integer = 5

    ' PPU Registers
    '---------------------------------------
    'Registers
    Public PPU_Control1 As Byte '$2000
    Public PPU_Control2 As Byte '$2001
    Public PPU_Status As Byte '$2002
    Public Loopy_V As Integer '$2006
    Public Loopy_T As Integer
    Public Loopy_X As Integer
    Public HScroll As Byte, VScroll As Integer
    'Misc
    Public TmpLatch As Byte, PPU_Latch As Byte
    Public SpriteAddress As Integer '$2003
    Public PPU_AddressIsHi As Boolean '$2005/$2006 Hi Toggle
    Public PatternTable As Integer
    Public NameTable As Integer

    ' Arrays
    '---------------------------------------
    'Video RAM
    Public VRAM(&H3FFF) As Byte
    Public SpriteRAM(&HFF) As Byte
    'Video Buffer
    Public Video_Buffer(256 * 241) As Integer
    Private Palette_32(31) As Integer
    'Screen rendering function
    Public Name_Table_Array(0 To 3, &H0 To &H3FF) As Byte
    Public Mirror(3) As Byte
    Public Palette(&HFF&) As Integer
    Public MirrorXor As Integer
    Public Sub Loopy_Scanline_Start(ByVal ScanLine As Long)
        If (PPU_Control2 And &H8) Or (PPU_Control2 And &H10) Then
            If ScanLine = 0 Then
                loopy_v = loopy_t
            Else
                loopy_v = (loopy_v And &HFBE0&) Or (loopy_t And &H41F&)
            End If
        End If
    End Sub
    Public Sub Loopy_Scanline_Next()
        If (PPU_Control2 And &H8) Or (PPU_Control2 And &H10) Then
            If (loopy_v And &H7000&) = &H7000& Then
                loopy_v = loopy_v And &H8FFF&
                If (loopy_v And &H3E0&) = &H3A0& Then
                    loopy_v = loopy_v Xor &H800&
                    loopy_v = loopy_v And &HFC1F&
                Else
                    If (loopy_v And &H3E0&) = &H3E0& Then
                        loopy_v = loopy_v And &HFC1F&
                    Else
                        loopy_v = loopy_v + &H20&
                    End If
                End If
            Else
                loopy_v = loopy_v + &H1000&
            End If
        End If
    End Sub
    Public Sub Render_ScanLine(ByVal ScanLine As Long)
        If (PPU_Control2 And &H8) = 0 And (PPU_Control2 And &H10) = 0 Or Not Render Then
            If ScanLine > SpriteRAM(0) + 8 Then PPU_Status = PPU_Status Or &H40
            Exit Sub
        End If

        If ScanLine > 240 Then Exit Sub

        For i As Integer = 0 To 31 Step 4
            Palette_32(i) = Palette(VRAM(i + &H3F00))
            Palette_32(i + 1) = Palette(VRAM(i + 1 + &H3F00))
            Palette_32(i + 2) = Palette(VRAM(i + 2 + &H3F00))
            Palette_32(i + 3) = Palette(VRAM(i + 3 + &H3F00))
        Next i
        If ScanLine < 240 Then
            For i As Integer = 256 * ScanLine To (256 * ScanLine) + 256 Step 4
                Video_Buffer(i) = Palette_32(16)
                Video_Buffer(i + 1) = Palette_32(16)
                Video_Buffer(i + 2) = Palette_32(16)
                Video_Buffer(i + 3) = Palette_32(16)
            Next i
        End If

        Loopy_Scanline_Start(ScanLine)
        Render_Sprites(ScanLine, False)
        Render_Background(ScanLine)
        Render_Sprites(ScanLine, True)
    End Sub
    Public Sub Render_Background(ByVal ScanLine As Long)
        Dim NameTable_Number As Byte
        Dim TileRow As Byte, TileYOffset As Integer
        Dim TileCounter As Integer
        Dim Color As Integer
        Dim TileIndex As Byte, Low_Byte As Byte, High_Byte As Byte
        Dim LookUp As Byte, Add_To_Color As Integer
        Dim Pixel As Integer
        Dim X As Integer, Aa As Integer
        Dim Offset As Integer
        Dim Virtual_Scanline As Integer
        Dim Attribute_Table As Long

        HScroll = (loopy_v And 31) * 8 + loopy_x
        VScroll = ((loopy_v \ 32 And 31) * 8 Or ((loopy_v \ &H1000&) And 7)) - ScanLine
        Loopy_Scanline_Next()

        If ScanLine = 239 Then PPU_Status = PPU_Status Or 128

        Virtual_Scanline = ScanLine + VScroll
        If Virtual_Scanline > 480 Then Virtual_Scanline -= 480
        TileRow = (Virtual_Scanline \ 8) Mod 30
        TileYOffset = Virtual_Scanline And 7

        NameTable = &H2000 + (loopy_v And &HFFF)
        NameTable_Number = (NameTable And &HC00) \ &H400

        Attribute_Table = &H3C0
        PatternTable = (PPU_Control1 And &H10) * &H100
        For TileCounter = HScroll \ 8 To 31
            TileIndex = Name_Table_Array(Mirror(NameTable_Number), TileCounter + TileRow * 32)

            X = TileCounter * 8 - HScroll + 7
            If X < 7 Then Offset = X Else Offset = 7
            X = X + ScanLine * 256
            LookUp = Name_Table_Array(Mirror(NameTable_Number), (&H3C0 + TileCounter \ 4 + (TileRow \ 4) * &H8))
            Select Case (TileCounter And 2) Or (TileRow And 2) * 2
                Case 0
                    Add_To_Color = LookUp * 4 And 12
                Case 2
                    Add_To_Color = LookUp And 12
                Case 4
                    Add_To_Color = LookUp \ 4 And 12
                Case 6
                    Add_To_Color = LookUp \ 16 And 12
            End Select
            Low_Byte = VRAM(PatternTable + TileIndex * 16 + TileYOffset)
            High_Byte = VRAM(PatternTable + TileIndex * 16 + 8 + TileYOffset)
            Aa = Low_Byte * 2048 + High_Byte * 8
            For Pixel = Offset To 0 Step -1
                Color = TLook(Aa + Pixel)
                If Color Mod 4 <> 0 Then Video_Buffer(X - Pixel) = Palette_32(Color Or Add_To_Color)
            Next Pixel
        Next TileCounter

        NameTable = NameTable Xor &H400&
        NameTable_Number = (NameTable And &HC00&) \ &H400&

        Attribute_Table = &H3C0
        For TileCounter = 0 To HScroll \ 8
            TileIndex = Name_Table_Array(Mirror(NameTable_Number), TileCounter + TileRow * 32)

            X = TileCounter * 8 + 256 - HScroll + 7
            If X > 255 Then Offset = X - 255 Else Offset = 0
            X = X + ScanLine * 256
            LookUp = Name_Table_Array(Mirror(NameTable_Number), (&H3C0 + TileCounter \ 4 + (TileRow \ 4) * &H8))
            Select Case (TileCounter And 2) Or (TileRow And 2) * 2
                Case 0
                    Add_To_Color = LookUp * 4 And 12
                Case 2
                    Add_To_Color = LookUp And 12
                Case 4
                    Add_To_Color = LookUp \ 4 And 12
                Case 6
                    Add_To_Color = LookUp \ 16 And 12
            End Select
            Low_Byte = VRAM(PatternTable + (TileIndex * 16) + TileYOffset)
            High_Byte = VRAM(PatternTable + (TileIndex * 16) + 8 + TileYOffset)
            Aa = Low_Byte * 2048 + High_Byte * 8
            For Pixel = 7 To Offset Step -1
                Color = TLook(Aa + Pixel)
                If Color Mod 4 <> 0 Then Video_Buffer(X - Pixel) = Palette_32(Color Or Add_To_Color)
            Next Pixel
        Next TileCounter
    End Sub
    Public Sub Render_Sprites(ByVal Scanline As Long, ByVal Drawn_In_Foreground As Boolean)
        Dim Current_Sprite_Address As Integer
        Dim Pixel_Color As Byte
        Dim Sprite_Offset As Integer
        Dim Index As Integer

        Dim SpriteX, SpriteY, SprPal As Integer
        Dim TileHeight As Integer = 8
        Dim TileIndex, Attr As Byte
        Dim Pattern_Table_Address As Integer
        Dim Drawn_In_Background, Vertical_Flip, Horizontal_Flip As Boolean

        Dim Low_Byte As Byte, High_Byte As Byte
        Dim Current_X_Pixel As Integer, Sprite_Scanline_To_Draw As Integer

        For Current_Sprite_Address = 252 To 0 Step -4
            Index = Current_Sprite_Address / 4

            SpriteY = SpriteRAM(Current_Sprite_Address) + 1
            TileIndex = SpriteRAM(Current_Sprite_Address + 1)
            Attr = SpriteRAM(Current_Sprite_Address + 2)
            SpriteX = SpriteRAM(Current_Sprite_Address + 3)

            If PPU_Control1 And 32 Then TileHeight = 16
            If PPU_Control1 And 8 Then Pattern_Table_Address = &H1000
            Drawn_In_Background = (Attr And 32)
            SprPal = 16 + (Attr And 3) * 4
            Vertical_Flip = Attr And 128
            Horizontal_Flip = Attr And 64

            If (Drawn_In_Background <> Drawn_In_Foreground) And (SpriteY <= Scanline) And ((SpriteY + TileHeight) > Scanline) Then
                If TileHeight = 8 Then
                    If Vertical_Flip = False Then
                        Sprite_Scanline_To_Draw = Scanline - SpriteY
                    Else 'Vertical Flip
                        Sprite_Scanline_To_Draw = SpriteY + 7 - Scanline
                    End If

                    Sprite_Offset = Pattern_Table_Address + (TileIndex * 16)
                    Low_Byte = VRAM(Sprite_Offset + Sprite_Scanline_To_Draw)
                    High_Byte = VRAM(Sprite_Offset + Sprite_Scanline_To_Draw + 8)

                    For Current_X_Pixel = 0 To 7
                        If Horizontal_Flip = True Then  'Horizontal Flip
                            Pixel_Color = SprPal + (((High_Byte And Power_Of_2(Current_X_Pixel)) \ Power_Of_2(Current_X_Pixel)) * 2) + ((Low_Byte And Power_Of_2(Current_X_Pixel)) \ Power_Of_2(Current_X_Pixel))
                        Else
                            Pixel_Color = SprPal + (((High_Byte And Power_Of_2(7 - Current_X_Pixel)) \ Power_Of_2(7 - Current_X_Pixel)) * 2) + ((Low_Byte And Power_Of_2(7 - Current_X_Pixel)) \ Power_Of_2(7 - Current_X_Pixel))
                        End If

                        If Pixel_Color Mod 4 <> 0 Then 'Adding Mod 4 <> 0 adds transparency to sprites
                            If (SpriteX + Current_X_Pixel) < 256 Then
                                Video_Buffer(Scanline * 256 + SpriteX + Current_X_Pixel) = Palette_32(SprPal Or Pixel_Color)
                                If Current_Sprite_Address = 0 Then PPU_Status = PPU_Status Or 64
                            End If
                        End If
                    Next Current_X_Pixel
                ElseIf TileHeight = 16 Then
                    If Vertical_Flip = False Then
                        Sprite_Scanline_To_Draw = Scanline - SpriteY
                    Else
                        Sprite_Scanline_To_Draw = SpriteY + 15 - Scanline
                    End If

                    If Sprite_Scanline_To_Draw < 8 Then
                        If TileIndex Mod 2 = 0 Then
                            Sprite_Offset = &H0 + TileIndex * 16
                        Else
                            Sprite_Offset = &H1000 + (TileIndex - 1) * 16
                        End If
                    Else
                        Sprite_Scanline_To_Draw = Sprite_Scanline_To_Draw - 8
                        If TileIndex Mod 2 = 0 Then
                            Sprite_Offset = &H0 + (TileIndex + 1) * 16
                        Else
                            Sprite_Offset = &H1000 + TileIndex * 16
                        End If
                    End If

                    Low_Byte = VRAM(Sprite_Offset + Sprite_Scanline_To_Draw)
                    High_Byte = VRAM(Sprite_Offset + Sprite_Scanline_To_Draw + 8)

                    For Current_X_Pixel = 0 To 7
                        If Horizontal_Flip = True Then  'Horizontal Flip
                            Pixel_Color = SprPal + (((High_Byte And Power_Of_2(Current_X_Pixel)) \ Power_Of_2(Current_X_Pixel)) * 2) + ((Low_Byte And Power_Of_2(Current_X_Pixel)) \ Power_Of_2(Current_X_Pixel))
                        Else
                            Pixel_Color = SprPal + (((High_Byte And Power_Of_2(7 - Current_X_Pixel)) \ Power_Of_2(7 - Current_X_Pixel)) * 2) + ((Low_Byte And Power_Of_2(7 - Current_X_Pixel)) \ Power_Of_2(7 - Current_X_Pixel))
                        End If

                        If Pixel_Color Mod 4 <> 0 Then 'Adding Mod 4 <> 0 adds transparency to sprites
                            If (SpriteX + Current_X_Pixel) < 256 Then
                                Video_Buffer(Scanline * 256 + SpriteX + Current_X_Pixel) = Palette_32(SprPal Or Pixel_Color)
                                If Current_Sprite_Address = 0 Then PPU_Status = PPU_Status Or 64
                            End If
                        End If
                    Next Current_X_Pixel
                End If
            End If
        Next Current_Sprite_Address
    End Sub
    Public Sub Do_Mirroring()
        MirrorXor = (((Mirroring + 1) Mod 3) * &H400&)
        If Mirroring = 0 Then 'H
            Mirror(0) = 0 : Mirror(1) = 0 : Mirror(2) = 1 : Mirror(3) = 1
        ElseIf Mirroring = 1 Then 'V
            Mirror(0) = 0 : Mirror(1) = 1 : Mirror(2) = 0 : Mirror(3) = 1
        ElseIf Mirroring = 2 Then '4L
            Mirror(0) = 0 : Mirror(1) = 0 : Mirror(2) = 0 : Mirror(3) = 0
        ElseIf Mirroring = 3 Then '4H
            Mirror(0) = 1 : Mirror(1) = 1 : Mirror(2) = 1 : Mirror(3) = 1
        ElseIf Mirroring = 4 Then '4
            Mirror(0) = 0 : Mirror(1) = 1 : Mirror(2) = 2 : Mirror(3) = 3
        End If
    End Sub
    Public Sub Blit(ByVal Buffer() As Integer, ByVal W As Integer, ByVal H As Integer, Optional ByVal MyZoom As Integer = 1, Optional ByVal Interpolation As Integer = 5)
        Dim bm1 As New Bitmap(W, H, Imaging.PixelFormat.Format32bppRgb)
        Dim bm2 As New Bitmap(W * MyZoom, H * MyZoom, Imaging.PixelFormat.Format32bppRgb)
        Dim GR As Graphics = Graphics.FromImage(bm2)
        GR.InterpolationMode = Interpolation
        Dim BitmapData1 As Imaging.BitmapData
        BitmapData1 = bm1.LockBits(New Rectangle(0, 0, W, H), Imaging.ImageLockMode.WriteOnly, Imaging.PixelFormat.Format32bppRgb)
        Dim Scan0 As IntPtr = BitmapData1.Scan0
        Runtime.InteropServices.Marshal.Copy(Buffer, 0, Scan0, W * H)
        bm1.UnlockBits(BitmapData1)
        GR.DrawImage(bm1, 0, 0, bm2.Width, bm2.Height)
        Form1.NesScreen.Image = bm2
    End Sub
End Module
